import math
from typing import List, Tuple, Union

import numpy as np


def heat_loss_u_values_pipe(
    inner_diameter: float,
    insulation_thicknesses: Union[float, List[float], np.ndarray] = None,
    conductivities_insulation: Union[float, List[float], np.ndarray] = 0.033,
    conductivity_subsoil: float = 2.3,
    depth: float = 1.0,
    h_surface: float = 15.4,
    pipe_distance: float = None,
) -> Tuple[float, float]:
    """
    Calculate the U_1 and U_2 heat loss values for a pipe based for either
    single- or multi-layer insulation. The heat loss calculation for two parallel pipes in the
    ground is based on the literature of Benny Böhm:
        - Original reference (paid article): B. Bohm, On transient heat losses from buried district
        heating pipes, International Journal of Energy Research 24, 1311 (2000))
        - Used in Master's degree: Jort de Boer, Optimization of a District Heating Network with the
        focus on heat loss, TU Delft Mechanical, Maritime and Materials Engineering, 2018,
        http://resolver.tudelft.nl/uuid:7be9fcdd-49e4-4e0c-b36c-69d8b713a874 (access to pdf)

    If the `insulation_thicknesses` is provided as a list, the length should be
    equal to the length of `conductivities_insulation`. If both are floats, a
    single layer of insulation is assumed.

    :inner_diameter:            Inner diameter of the pipes [m]
    :insulation_thicknesses:    Thicknesses of the insulation [m]
                                Default of None means a thickness of 0.5 * inner diameter.
    :conductivities_insulation: Thermal conductivities of the insulation layers [W/m/K]
    :conductivity_subsoil:      Subsoil thermal conductivity [W/m/K]
    :h_surface:                 Heat transfer coefficient at surface [W/m^2/K]
    :param depth:               Depth of outer top of the pipeline [m]
    :param pipe_distance:       Distance between pipeline feed and return pipeline centers [m].
                                Default of None means 2 * outer diameter

    :return: U-values (U_1 / U_2) for heat losses of pipes [W/(m*K)]
    """

    if insulation_thicknesses is None:
        insulation_thicknesses = 0.5 * inner_diameter

    if not isinstance(insulation_thicknesses, type(conductivities_insulation)):
        raise Exception("Insulation thicknesses and conductivities should have the same type.")

    if hasattr(insulation_thicknesses, "__iter__"):
        if not len(insulation_thicknesses) == len(conductivities_insulation):
            raise Exception(
                "Number of insulation thicknesses should match number of conductivities"
            )
        insulation_thicknesses = np.array(insulation_thicknesses)
        conductivities_insulation = np.array(conductivities_insulation)
    else:
        insulation_thicknesses = np.array([insulation_thicknesses])
        conductivities_insulation = np.array([conductivities_insulation])

    diam_inner = inner_diameter
    diam_outer = diam_inner + 2 * sum(insulation_thicknesses)
    if pipe_distance is None:
        pipe_distance = 2 * diam_outer
    depth_center = depth + 0.5 * diam_outer

    # NOTE: We neglect the heat resistance due to convection inside the pipe,
    # i.e. we assume perfect mixing, or that this resistance is much lower
    # than the resistance of the outer insulation layers.

    # Heat resistance of the subsoil
    r_subsoil = 1 / (2 * math.pi * conductivity_subsoil) * math.log(4.0 * depth_center / diam_outer)

    # Heat resistance due to insulation
    outer_diameters = diam_inner + 2.0 * np.cumsum(insulation_thicknesses)
    inner_diameters = np.array([inner_diameter, *outer_diameters[:-1]])
    r_ins = sum(
        np.log(outer_diameters / inner_diameters) / (2.0 * math.pi * conductivities_insulation)
    )

    # Heat resistance due to neighboring pipeline
    r_m = (
        1
        / (4 * math.pi * conductivity_subsoil)
        * math.log(1 + (2 * depth_center / pipe_distance) ** 2)
    )

    u_1 = (r_subsoil + r_ins) / ((r_subsoil + r_ins) ** 2 - r_m**2)
    u_2 = r_m / ((r_subsoil + r_ins) ** 2 - r_m**2)

    return u_1, u_2



import logging

from rtctools.optimization.timeseries import Timeseries
from rtctools.util import run_optimization_problem

from . import __version__


def run_heat_network_optimization(heat_class, qht_class, *args, log_level=logging.INFO, **kwargs):
    logger = logging.getLogger("rtctools_heat_network")
    logger.setLevel(log_level)
    logger.info(f"Using RTC-Tools Heat Network {__version__}.")

    heat_problem = run_optimization_problem(heat_class, *args, log_level=log_level, **kwargs)
    results = heat_problem.extract_results()
    times = heat_problem.times()

    directions = {}

    for p in heat_problem.hot_pipes:
        heat_in = results[p + ".HeatIn.Heat"]
        heat_out = results[p + ".HeatOut.Heat"]

        if not heat_problem.parameters(0)[p + ".disconnectable"]:
            # Flow direction is directly related to the sign of the heat
            direction_pipe = (heat_in >= 0.0).astype(int) * 2 - 1
        elif heat_problem.parameters(0)[p + ".disconnectable"]:
            direction_pipe = (heat_in >= 0.0).astype(int) * 2 - 1
            # Disconnect a pipe when the heat entering the component is only used
            # to account for its heat losses. There are three cases in which this
            # can happen.
            direction_pipe[((heat_in > 0.0) & (heat_out < 0.0))] = 0
            direction_pipe[((heat_in < 0.0) & (heat_out > 0.0))] = 0
            direction_pipe[((heat_in == 0.0) | (heat_out == 0.0))] = 0
        directions[p] = Timeseries(times, direction_pipe)

        # NOTE: The assumption is that the orientation of the cold pipes is such that the flow
        # is always in the same direction as its "hot" pipe companion.
        cold_pipe = heat_problem.hot_to_cold_pipe(p)
        directions[cold_pipe] = directions[p]

    for v in heat_problem.heat_network_components.get("check_valve", []):
        status_valve = (results[f"{v}__status_var"]).round().astype(int)
        directions[v] = Timeseries(times, status_valve)

    for v in heat_problem.heat_network_components.get("control_valve", []):
        directions_valve = (results[f"{v}__flow_direct_var"]).round().astype(int) * 2 - 1
        directions[v] = Timeseries(times, directions_valve)

    buffer_target_discharges = {}
    parameters = heat_problem.parameters(0)

    for b in heat_problem.heat_network_components.get("buffer", []):
        cp = parameters[f"{b}.cp"]
        rho = parameters[f"{b}.rho"]
        heat_flow_rate_to_discharge = 1 / (cp * rho * parameters[f"{b}.dT"])
        buffer_target_discharges[b] = Timeseries(
            times, results[f"{b}.Heat_buffer"] * heat_flow_rate_to_discharge
        )

    qth_problem = run_optimization_problem(
        qht_class,
        *args,
        log_level=log_level,
        flow_directions=directions,
        buffer_target_discharges=buffer_target_discharges,
        **kwargs,
    )

    return heat_problem, qth_problem



from dataclasses import dataclass

from rtctools_heat_network.esdl.asset_to_component_base import _AssetToComponentBase
from rtctools_heat_network.pipe_class import PipeClass


@dataclass(frozen=True)
class EDRPipeClass(PipeClass):
    xml_string: str

    @classmethod
    def from_edr_class(cls, name, edr_class_name, maximum_velocity):
        if not hasattr(EDRPipeClass, "._edr_pipes"):
            # TODO: Currently using private API of RTC-Tools Heat Network.
            # Make this functionality part of public API?
            EDRPipeClass._edr_pipes = _AssetToComponentBase()._edr_pipes

        edr_class = EDRPipeClass._edr_pipes[edr_class_name]
        diameter = edr_class["inner_diameter"]
        u_1 = edr_class["u_1"]
        u_2 = edr_class["u_2"]
        investment_costs = edr_class["investment_costs"]
        xml_string = edr_class["xml_string"]

        return EDRPipeClass(
            name, diameter, maximum_velocity, (u_1, u_2), investment_costs, xml_string
        )



from dataclasses import dataclass


@dataclass(frozen=True)
class DemandInsulationClass:
    name_insulation_level: str
    name_demand: str
    minimum_temperature_deg: float
    demand_scaling_factor: float  # value between 0.0 and 1.0
    investment_cost_euro: float



import itertools
import logging
from abc import abstractmethod
from typing import Dict, Union

import casadi as ca

import pymoca
from pymoca.backends.casadi.model import Model as _Model

from rtctools._internal.alias_tools import AliasDict
from rtctools._internal.caching import cached
from rtctools.optimization.optimization_problem import OptimizationProblem

from . import ConstantInput, ControlInput, Model, SymbolicParameter, Variable


logger = logging.getLogger("rtctools_heat_network")


class PyCMLMixin(OptimizationProblem):
    def __init__(self, *args, **kwargs):
        logger.debug("Using pymoca {}.".format(pymoca.__version__))

        pycml_model = self.pycml_model()

        self.__flattened_model = pycml_model.flatten()

        self.__pymoca_model = _Model()
        for v in self.__flattened_model.variables.values():
            if isinstance(v, SymbolicParameter):
                self.__pymoca_model.parameters.append(v)
            elif isinstance(v, (ControlInput, ConstantInput)):
                self.__pymoca_model.inputs.append(v)
            elif isinstance(v, Variable) and v.has_derivative:
                self.__pymoca_model.states.append(v)
                self.__pymoca_model.der_states.append(v.der())
            else:
                self.__pymoca_model.alg_states.append(v)

        self.__pymoca_model.equations = self.__flattened_model.equations
        self.__pymoca_model.initial_equations = self.__flattened_model.initial_equations
        self.__pymoca_model.simplify(self.compiler_options())

        if (
            len(self.__flattened_model.inequalities) > 0
            or len(self.__flattened_model.initial_inequalities) > 0
        ):
            raise NotImplementedError("Inequalities are not supported yet")

        # Note that we do not pass the numeric parameters to the Pymoca model
        # in their entirety. That way we can avoid making useless Variable
        # instances, as the parameters do not appear in any equations anyway.
        self.__parameters = {
            k: v
            for k, v in self.__flattened_model.numeric_parameters.items()
            if not isinstance(v, str)
        }
        self.__string_parameters = {
            k: v for k, v in self.__flattened_model.numeric_parameters.items() if isinstance(v, str)
        }

        # Extract the CasADi MX variables used in the model
        self.__mx = {}
        self.__mx["time"] = [self.__pymoca_model.time]
        self.__mx["states"] = [v.symbol for v in self.__pymoca_model.states]
        self.__mx["derivatives"] = [v.symbol for v in self.__pymoca_model.der_states]
        self.__mx["algebraics"] = [v.symbol for v in self.__pymoca_model.alg_states]
        self.__mx["parameters"] = [v.symbol for v in self.__pymoca_model.parameters]
        self.__mx["string_parameters"] = [
            v.name
            for v in (*self.__pymoca_model.string_parameters, *self.__pymoca_model.string_constants)
        ]
        self.__mx["control_inputs"] = []
        self.__mx["constant_inputs"] = []
        self.__mx["lookup_tables"] = []

        for v in self.__pymoca_model.inputs:
            if v.symbol.name() in self.__pymoca_model.delay_states:
                raise NotImplementedError("Delays are not supported yet")
            else:
                if v.symbol.name() in kwargs.get("lookup_tables", []):
                    raise NotImplementedError()
                elif v.fixed:
                    self.__mx["constant_inputs"].append(v.symbol)
                else:
                    self.__mx["control_inputs"].append(v.symbol)

        # Initialize nominals and types
        # These are not in @cached dictionary properties for backwards compatibility.
        self.__python_types = AliasDict(self.alias_relation)
        for v in itertools.chain(
            self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs
        ):
            self.__python_types[v.symbol.name()] = v.python_type

        # Initialize dae, initial residuals, as well as delay arguments
        # These are not in @cached dictionary properties so that we need to create the list
        # of function arguments only once.
        variable_lists = ["states", "der_states", "alg_states", "inputs", "constants", "parameters"]
        function_arguments = [self.__pymoca_model.time] + [
            ca.veccat(*[v.symbol for v in getattr(self.__pymoca_model, variable_list)])
            for variable_list in variable_lists
        ]

        self.__dae_residual = self.__pymoca_model.dae_residual_function(*function_arguments)
        if self.__dae_residual is None:
            self.__dae_residual = ca.MX()

        self.__initial_residual = self.__pymoca_model.initial_residual_function(*function_arguments)
        if self.__initial_residual is None:
            self.__initial_residual = ca.MX()

        super().__init__(*args, **kwargs)

    @cached
    def compiler_options(self) -> Dict[str, Union[str, bool]]:
        """
        Subclasses can configure the `pymoca <http://github.com/pymoca/pymoca>`_
        compiler options here.

        :returns: A dictionary of pymoca compiler options. See the pymoca documentation
                  for details.
        """
        compiler_options = {}
        compiler_options["detect_aliases"] = True
        compiler_options["replace_parameter_expressions"] = True
        compiler_options["allow_derivative_aliases"] = False
        return compiler_options

    @property
    def dae_residual(self):
        return self.__dae_residual

    @property
    def dae_variables(self):
        return self.__mx

    @property
    def output_variables(self):
        output_variables = super().output_variables.copy()
        output_variables.extend([ca.MX.sym(variable) for variable in self.__pymoca_model.outputs])
        output_variables.extend(self.__mx["control_inputs"])
        return output_variables

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)
        parameters.update(self.__parameters)
        parameters.update({v.name: v.value for v in self.__pymoca_model.parameters})
        return parameters

    def string_parameters(self, ensemble_member):
        parameters = super().string_parameters(ensemble_member)
        parameters.update(self.__string_parameters)
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_parameters})
        parameters.update({v.name: v.value for v in self.__pymoca_model.string_constants})
        return parameters

    @property
    def initial_residual(self):
        return self.__initial_residual

    def bounds(self):
        bounds = super().bounds()

        for v in itertools.chain(
            self.__pymoca_model.states, self.__pymoca_model.alg_states, self.__pymoca_model.inputs
        ):
            sym_name = v.symbol.name()

            try:
                bounds[sym_name] = self.merge_bounds(bounds[sym_name], (v.min, v.max))
            except KeyError:
                if self.__python_types.get(sym_name, float) == bool:
                    bounds[sym_name] = (max(0, v.min), min(1, v.max))
                else:
                    bounds[sym_name] = (v.min, v.max)

        return bounds

    def variable_is_discrete(self, variable):
        return self.__python_types.get(variable, float) != float

    @property
    def alias_relation(self):
        return self.__pymoca_model.alias_relation

    def variable_nominal(self, variable):
        try:
            return self.__nominals[variable]
        except AttributeError:
            self.__nominals = AliasDict(self.alias_relation)

            # Iterate over nominalizable states
            for v in itertools.chain(
                self.__pymoca_model.states,
                self.__pymoca_model.alg_states,
                self.__pymoca_model.inputs,
            ):
                sym_name = v.symbol.name()
                nominal = v.nominal
                if nominal == 0.0:
                    nominal = 1.0
                self.__nominals[sym_name] = nominal

            return self.variable_nominal(variable)
        except KeyError:
            return super().variable_nominal(variable)

    @abstractmethod
    def pycml_model(self) -> Model:
        raise NotImplementedError



from typing import Dict, List, Tuple

import casadi as ca

import numpy as np

from rtctools._internal.alias_tools import AliasDict
from rtctools.optimization.goal_programming_mixin_base import Goal
from rtctools.optimization.timeseries import Timeseries

from rtctools_heat_network.qth_mixin import HeadLossOption, QTHMixin


class BufferTargetDischargeGoal(Goal):
    def __init__(
        self,
        buffer: str,
        time: float,
        target: float,
        function_range: Tuple[float, float],
        *args,
        priority=2,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.buffer = buffer

        self.priority = priority

        self._time = time
        self.target_min = target
        self.target_max = target

        self.function_range = function_range
        self.function_nominal = max(np.abs(function_range)) / 2.0

    def function(self, optimization_problem, ensemble_member):
        return optimization_problem.state_at(
            f"{self.buffer}.Q_hot_pipe", self._time, ensemble_member
        )


class QTHLoopMixin(QTHMixin):
    """
    Alternative to QTHMixin when the assumptions of sufficient hydraulic
    control do not hold for a certain heat network. This Mixin runs 1 timestep
    at a time, and solves with the `CQ2_EQUALITY` head loss option to get to a
    hydraulically feasible result for such heat networks.
    More precisely, the Mixin loops over the time horizon optimizing two
    timesteps at each iteration where the solution of first timestep is de facto
    fixed, using the result of the previous iteration of the loop. As IPOPT does
    not tolerate trivial constraints, we fix the minimum amount of necessary information.
    That is, the temperature of all the sources and buffers and all the discharges
    within the network.
    Moreover, tote that some values and defaults of QTHMixin are overruled that:

    - Do not make sense when optimizing a single time step

    - Can more readily lead to infeasible problems when not optimizing the
      whole horizon at once.

    Not optimizing over the whole horizon also means that some (former)
    control inputs now need to be prescribed, most notably the discharge to
    and from buffers. It is recommendable that these goals are added inbetween
    the goals related to the demands and the ones related to the sources.
    See :class:`buffer_target_discharges` and :class:`BufferTargetDischargeGoal`,
    and the helper method :meth:`buffer_target_discharge_goals`
    """

    def __init__(self, *args, flow_directions=None, buffer_target_discharges=None, **kwargs):
        super().__init__(*args, flow_directions=flow_directions, **kwargs)

        self.__buffer_target_discharges = buffer_target_discharges

        self.__extended_history = None
        self.__expose_all_results = False
        self.__all_results = []

    def pre(self):
        super().pre()

        if self.__tstep == 0:
            homotopy_options = self.homotopy_options()
            self.__homotopy_theta = homotopy_options["homotopy_parameter"]
            self.__extended_history = None
            self.__expose_all_results = False
            self.__all_results = []

    @property
    def buffer_target_discharges(self) -> Dict[str, Timeseries]:
        return self.__buffer_target_discharges

    def buffer_target_discharge_goals(
        self, priority=2, function_range_max_fac=2.0
    ) -> List[BufferTargetDischargeGoal]:
        goals = []
        times = self.times()

        for b, t in self.buffer_target_discharges.items():
            function_range_max = max(np.abs(t.values)) * function_range_max_fac
            target = self.interpolate(times[1], t.times, t.values)
            function_range = (-function_range_max, function_range_max)
            goals.append(
                BufferTargetDischargeGoal(b, times[1], target, function_range, priority=priority)
            )

        return goals

    def goal_programming_options(self):
        options = super().goal_programming_options()

        # The optimization problems only have two timesteps, and issues due to
        # overconstraining it are more likely than leaving too much freedom.
        options["keep_soft_constraints"] = True

        return options

    def heat_network_options(self):
        """
        Returns a dictionary of heat network specific options.

        See :py:meth:`QTHMixin.heat_network_options` for all options. When
        inheriting from QTHLoopMixin, some defaults are changed:

        - `head_loss_option` is CQ2_EQUALITY
        - `max_t_der_bidirect_pipe` is False
        The latter is due to the fact that the loop optimization does not
        allow to control the temperatures within the buffer.
        """

        options = super().heat_network_options()
        options["head_loss_option"] = HeadLossOption.CQ2_EQUALITY
        options["max_t_der_bidirect_pipe"] = False
        return options

    def homotopy_options(self):
        options = super().homotopy_options()

        if self.__tstep > 0:
            options["theta_start"] = 1.0

        return options

    def times(self, variable=None):
        times = super().times(variable)
        if self.__expose_all_results:
            return times[: self.__tstep + 2]
        else:
            return times[self.__tstep : self.__tstep + 2]

    def parameters(self, ensemble_member):
        parameters = super().parameters(ensemble_member)

        if self.__tstep > 0:
            parameters[self.__homotopy_theta] = 1.0

        return parameters

    def __check_goals(self):
        if not len(self.times()) == 2:
            raise Exception("QTHLoopMixin is only allowed to run with 2 time steps")

        buffer_goals = {
            g.buffer: g for g in self.goals() if isinstance(g, BufferTargetDischargeGoal)
        }

        for b in self.heat_network_topology.buffers:
            if b not in buffer_goals:
                raise Exception(f"Buffer {b} is missing a corresponding BufferTargetDischargeGoal")

    def __process_results(self):
        """
        We do two things:

        1. We append the t0 of the current solution to the history (it becomes "t-1"
           for the next run)

        2. We store _part of_ the t1 results of the current solution, which will
           become the next t0 history

        Note that we don't want to mess with the original history AliasDict to avoid
        unexpected side-effects, which is why we make a deepcopy.
        """

        if self.__tstep == 0:
            assert self.__extended_history is None

            self.__extended_history = []

            for ensemble_member in range(self.ensemble_size):
                history = self.history(ensemble_member)
                extended_history = history.copy()

                # Make the copy "deep" by reinstantiating the Timeseries
                for k, v in extended_history.items():
                    extended_history[k] = Timeseries(v.times, v.values)

                self.__extended_history.append(extended_history)

        times = self.times()
        self.__prev_t1_solver_output = self.solver_output.copy()
        self.__prev_t1_solution_dict = []

        if self.__all_results is None:
            self.__all_results = []

        for ensemble_member in range(self.ensemble_size):
            results = self.extract_results(ensemble_member)

            if len(self.__all_results) <= ensemble_member:
                self.__all_results.append(AliasDict(self.alias_relation))
            cur_all_results = self.__all_results[ensemble_member]

            extended_history = self.__extended_history[ensemble_member]
            prev_solution = {}

            for k, v in results.items():
                if len(v) == 1:
                    # Initial derivative type of stuff
                    continue

                if self.__tstep == 0:
                    extended_history[k] = Timeseries(times, [v[0], np.nan])
                else:
                    ts = extended_history[k]
                    extended_history[k] = Timeseries(
                        [*ts.times[:-1], *times], [*ts.values[:-1], v[0], np.nan]
                    )

                if self.__tstep == 0:
                    cur_all_results[k] = [v[0], v[1]]
                else:
                    cur_all_results[k].append(v[1])

                prev_solution[k] = v[1]

            self.__prev_t1_solution_dict.append(prev_solution)

    def history(self, ensemble_member):
        if self.__tstep > 0:
            return self.__extended_history[ensemble_member]
        else:
            return super().history(ensemble_member)

    def seed(self, ensemble_member):
        seed = super().seed(ensemble_member)

        if self.__tstep > 0:
            times = self.times()

            for k, v in self.__prev_t1_solution_dict[ensemble_member].items():
                seed[k] = Timeseries(times, [v, v])

        return seed

    def transcribe(self):
        discrete, lbx, ubx, lbg, ubg, x0, nlp = super().transcribe()

        if self.__tstep > 0:
            # We overrule here instead of in bounds(), because bounds() does
            # not support per-ensemble-member bounds. The collocation indices
            # are private for now, but will become part of the public API soon.
            parameters = self.parameters(0)
            assert parameters[self.__homotopy_theta] == 1.0

            lb = np.full_like(lbx, -np.inf)
            ub = np.full_like(ubx, np.inf)

            fix_value_variables = set()

            for s in self.heat_network_components["source"]:
                fix_value_variables.add(self.alias_relation.canonical_signed(f"{s}.QTHOut.T")[0])

            for b in self.heat_network_components.get("buffer", []):
                fix_value_variables.add(self.alias_relation.canonical_signed(f"{b}.T_hot_tank")[0])
                fix_value_variables.add(self.alias_relation.canonical_signed(f"{b}.T_cold_tank")[0])

            for p in self.heat_network_components["pipe"]:
                fix_value_variables.add(self.alias_relation.canonical_signed(f"{p}.Q")[0])

            previous_indices = self.__previous_indices
            current_indices = self._CollocatedIntegratedOptimizationProblem__indices_as_lists

            for ensemble_member in range(self.ensemble_size):
                for v in fix_value_variables:
                    cur_inds = current_indices[ensemble_member][v]
                    prev_inds = previous_indices[ensemble_member][v]

                    assert len(cur_inds) == 2
                    assert len(prev_inds) == 2

                    ub[cur_inds[0]] = lb[cur_inds[0]] = self.__prev_t1_solver_output[prev_inds[1]]

            lbx = np.maximum(lbx, lb)
            ubx = np.minimum(ubx, ub)

            # Sometimes rounding errors can change the ulp, leading to lbx
            # becomes slightly (~1E-15) larger than ubx. Fix by setting
            # equality entries explicitly.
            inds = lb == ub
            lbx[inds] = ubx[inds] = lb[inds]

            assert np.all(lbx <= ubx)

        self.__previous_indices = self._CollocatedIntegratedOptimizationProblem__indices_as_lists

        return discrete, lbx, ubx, lbg, ubg, x0, nlp

    def extract_results(self, ensemble_member=0):
        if self.__expose_all_results:
            return self.__all_results[ensemble_member]
        else:
            return super().extract_results(ensemble_member)

    def optimize(
        self,
        preprocessing: bool = True,
        postprocessing: bool = True,
        log_solver_failure_as_error: bool = True,
    ) -> bool:
        self.__tstep = 0

        if preprocessing:
            self.pre()

        n_times = len(super().times())

        self.__check_goals()

        for self.__tstep in range(n_times - 1):
            success = super().optimize(preprocessing=preprocessing, postprocessing=False)
            self.__process_results()

            if not success:
                break

            if self.__tstep == 0:
                # We will get rid of the initial residual from the next run
                # onwards, so clear the cache.
                self.clear_transcription_cache()

        for results in self.__all_results:
            for k, v in results.items():
                results[k] = np.array(v)

        self.__expose_all_results = True

        if postprocessing:
            self.post()

    @property
    def initial_residual(self):
        if self.__tstep > 0:
            return ca.MX()
        else:
            return super().initial_residual



from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem


class TestHEX(TestCase):
    def test_heat_exchanger(self):
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import (
            HeatProblem,
        )

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        prim_heat = results["HeatExchange_39ed.Primary_heat"]
        sec_heat = results["HeatExchange_39ed.Secondary_heat"]
        prim_q = results["HeatExchange_39ed.Primary.HeatIn.Q"]
        sec_q = results["HeatExchange_39ed.Secondary.HeatOut.Q"]
        disabled = results["HeatExchange_39ed__disabled"]

        # Values used in non_storage_component.py
        cp = 4200.0
        rho = 988.0

        # We check the energy converted betweeen the commodities
        # 0.9 efficiency specified in esdl
        np.testing.assert_allclose(prim_heat * 0.9, sec_heat)
        # We check the energy being properly linked to the flows, this should already be satisfied
        # by the multiple commodity test but we do it anyway.
        np.testing.assert_allclose(prim_heat, prim_q * cp * rho * 40.0)
        # Note that we are not testing the last element as we exploit the last timestep for
        # checking the disabled boolean and the assert statement doesn't work for a difference of
        # zero
        np.testing.assert_array_less(-np.abs(sec_q[:-1] * cp * rho * 30.0 - sec_heat[:-1]), 1.0e-6)
        np.testing.assert_allclose(prim_heat[-1], 0.0)
        np.testing.assert_allclose(disabled[-1], 1.0)
        np.testing.assert_allclose(disabled[:-1], 0.0)


class TestHP(TestCase):
    def test_heat_pump(self):
        import models.heatpump.src.run_heat_pump as run_heat_pump
        from models.heatpump.src.run_heat_pump import (
            HeatProblem,
        )

        base_folder = Path(run_heat_pump.__file__).resolve().parent.parent

        solution = run_optimization_problem(HeatProblem, base_folder=base_folder)

        results = solution.extract_results()

        prim_heat = results["GenericConversion_3d3f.Primary_heat"]
        sec_heat = results["GenericConversion_3d3f.Secondary_heat"]
        power_elec = results["GenericConversion_3d3f.Power_elec"]
        prim_q = results["GenericConversion_3d3f.Primary.HeatIn.Q"]
        sec_q = results["GenericConversion_3d3f.Secondary.HeatOut.Q"]

        # TODO: we should also check if heatdemand target is matched
        # TODO: check if the primary source utilisisation is maximised and secondary minimised

        # Values used in non_storage_component.py
        cp = 4200.0
        rho = 988.0

        # We check the energy converted betweeen the commodities
        # 0.9 efficiency specified in esdl
        np.testing.assert_allclose(power_elec * 4.0, sec_heat)
        np.testing.assert_allclose(power_elec + prim_heat, sec_heat)
        # We check the energy being properly linked to the flows, this should already be satisfied
        # by the multiple commodity test but we do it anyway.
        np.testing.assert_allclose(prim_heat, prim_q * cp * rho * 30.0)
        np.testing.assert_array_less(sec_q * cp * rho * 40.0, sec_heat)



from numpy import nan

from rtctools_heat_network.pycml import Variable

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricitySource(ElectricityComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_source"
        self.meret_place = 1

        self.price = nan

        self.add_variable(ElectricityPort, "ElectricityOut")
        self.add_variable(Variable, "Electricity_source", min=0.0)

        self.add_equation((self.ElectricityOut.Power - self.Electricity_source))



from rtctools_heat_network.pycml import Variable

from .qth_two_port import QTHTwoPort


class _NonStorageComponent(QTHTwoPort):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "Q")

        self.add_variable(Variable, "H_in")
        self.add_variable(Variable, "H_out")

        self.add_equation(self.QTHIn.Q - self.Q)
        self.add_equation(self.QTHOut.Q - self.QTHIn.Q)

        self.add_equation(self.QTHIn.H - self.H_in)
        self.add_equation(self.QTHOut.H - self.H_out)



import logging
import os
import uuid
from pathlib import Path

import esdl

import numpy as np

import pandas as pd

from rtctools_heat_network.esdl.edr_pipe_class import EDRPipeClass
from rtctools_heat_network.heat_mixin import HeatMixin
from rtctools_heat_network.workflows.utils.helpers import _sort_numbered

logger = logging.getLogger("rtctools_heat_network")


class ScenarioOutput(HeatMixin):
    __optimized_energy_system_handler = None

    def get_optimized_esh(self):
        return self.__optimized_energy_system_handler

    def _write_html_output(self, template_name="mpc_buffer_sizing_output"):
        from jinja2 import Environment, FileSystemLoader

        assert self.ensemble_size == 1

        results = self.extract_results()
        parameters = self.parameters(0)

        # Format the priority results
        priority_results = [
            dict(
                number=number,
                success=success,
                pretty_time=f"{int(seconds // 60):02d}:{seconds % 60:06.3f}",
                objective_value=objective_value,
                return_status=stats["return_status"],
                secondary_return_status=stats.get("secondary_return_status", ""),
            )
            for (
                number,
                seconds,
                success,
                objective_value,
                stats,
            ) in self._priorities_output
        ]

        # Format the buffer results
        results_buffers_placed = {}
        results_buffers_size = {}
        results_sources_placed = {}
        results_sources_size = {}
        results_max_charging_rate = {}
        results_max_discharging_rate = {}

        for buffer in _sort_numbered(self.heat_network_components.get("buffer", [])):
            if buffer in self._minimize_size_buffers:
                max_size_var = self._max_buffer_heat_map[buffer]
                results_buffers_size[buffer] = float(results[max_size_var][0]) / (
                    parameters[f"{buffer}.cp"]
                    * parameters[f"{buffer}.rho"]
                    * (parameters[f"{buffer}.T_supply"] - parameters[f"{buffer}.T_return"])
                )
            else:
                results_buffers_size[buffer] = "-"

            if buffer in self._optional_buffers:
                buffer_placement_var = self._buffer_placement_map[buffer]
                results_buffers_placed[buffer] = np.round(results[buffer_placement_var][0]) == 1.0
            else:
                results_buffers_placed[buffer] = "-"

            (_, hot_orient), _ = self.heat_network_topology.buffers[buffer]
            q = hot_orient * results[f"{buffer}.HeatIn.Q"]
            inds_charging = q > 0
            inds_discharging = q < 0

            results_max_charging_rate[buffer] = max(q[inds_charging]) if any(inds_charging) else 0.0
            results_max_discharging_rate[buffer] = (
                max(-1 * q[inds_discharging]) if any(inds_discharging) else 0.0
            )

        buffer_results = [
            dict(
                name=buffer,
                tune_size=buffer in self._minimize_size_buffers,
                tune_placement=buffer in self._optional_buffers,
                maximum_size=self._override_buffer_size[buffer],
                result_placed=results_buffers_placed[buffer],
                result_size=results_buffers_size[buffer],
                max_charging_rate=results_max_charging_rate[buffer],
                max_discharging_rate=results_max_discharging_rate[buffer],
            )
            for buffer in self.heat_network_components.get("buffer", [])
        ]

        for source in _sort_numbered(self.heat_network_components["source"]):
            if source in self._minimize_size_sources:
                max_size_var = self._max_source_heat_map[source]
                results_sources_size[source] = float(results[max_size_var][0]) / 10.0**3
            else:
                results_sources_size[source] = "-"

            if source in self._optional_sources:
                source_placement_var = self._source_placement_map[source]
                results_sources_placed[source] = np.round(results[source_placement_var][0]) == 1.0
            else:
                results_sources_placed[source] = "-"

        source_results = [
            dict(
                name=source,
                tune_size=source in self._minimize_size_sources,
                tune_placement=source in self._optional_sources,
                maximum_size=self._override_max_power[source],
                result_placed=results_sources_placed[source],
                result_size=results_sources_size[source],
            )
            for source in self.heat_network_components["source"]
        ]

        # Format the pipe results
        # Note that we do not distinguish between routing and sizing
        # internally, but for the sake of the output we do.
        pipe_results = []

        for p in _sort_numbered(self.hot_pipes):
            pipe_classes = self.pipe_classes(p)
            tune_routing = len([pc for pc in pipe_classes if pc.inner_diameter == 0.0]) == 1
            inner_diameter = parameters[f"{p}.diameter"]
            asset = next(a for a in self.esdl_assets.values() if a.name == p)
            esdl_diameter = asset.attributes["innerDiameter"]

            if len(pipe_classes) <= 1:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "-"
                result_dn_size = "-"
            elif len(pipe_classes) == 2 and tune_routing:
                tune_size = False
                min_dn_size = "-"
                max_dn_size = "-"
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_dn_size = "-"
            else:
                sorted_pipe_classes = sorted(
                    [pc for pc in pipe_classes if pc.inner_diameter > 0],
                    key=lambda pc: pc.inner_diameter,
                )

                tune_size = True
                min_dn_size = sorted_pipe_classes[0].name
                max_dn_size = sorted_pipe_classes[-1].name
                result_placed = "Yes" if inner_diameter > 0 else "No"
                result_pipe_class = self.get_optimized_pipe_class(p)
                result_dn_size = (
                    result_pipe_class.name if result_pipe_class is not None else inner_diameter
                )

            pipe_results.append(
                dict(
                    name=p,
                    tune_size=tune_size,
                    tune_routing=tune_routing,
                    esdl_diameter=esdl_diameter,
                    min_dn_size=min_dn_size,
                    max_dn_size=max_dn_size,
                    result_placed=result_placed,
                    result_dn_size=result_dn_size,
                )
            )

        input_csv_tables = {
            os.path.basename(x): pd.read_csv(x).to_dict("records")
            for x in self._csv_input_parameter_files
        }

        # Actually write out the html file based on the template
        templates_dir = Path(__file__).parent / "templates"

        env = Environment(loader=FileSystemLoader(templates_dir))
        template = env.get_template(template_name + ".html")

        os.makedirs(self._html_output_dir, exist_ok=True)

        filename = self._html_output_dir / (template_name + ".html")

        with open(filename, "w", encoding="utf-8") as fh:
            fh.write(
                template.render(
                    buffer_results=buffer_results,
                    source_results=source_results,
                    pipe_results=pipe_results,
                    priority_results=priority_results,
                    input_csv_tables=input_csv_tables,
                )
            )

    def _write_updated_esdl(self, db_profiles=False, optimizer_sim=True):
        from esdl.esdl_handler import EnergySystemHandler
        from rtctools_heat_network.esdl.esdl_mixin import _RunInfoReader

        results = self.extract_results()
        parameters = self.parameters(0)

        esh = EnergySystemHandler()
        if self.esdl_string is None:
            run_info = _RunInfoReader(self.esdl_run_info_path)
            energy_system = esh.load_file(str(run_info.esdl_file))
        else:
            energy_system = esh.load_from_string(self.esdl_string)

        energy_system.id = str(uuid.uuid4())
        if optimizer_sim:
            energy_system.name = energy_system.name + "_GrowOptimized"
        else:
            energy_system.name = energy_system.name + "_Simulation"

        def _name_to_asset(name):
            return next(
                (x for x in energy_system.eAllContents() if hasattr(x, "name") and x.name == name)
            )

        # ------------------------------------------------------------------------------------------
        # KPIs
        # General cost breakdowns
        # ------------------------------------------------------------------------------------------
        kpis_top_level = esdl.KPIs(id=str(uuid.uuid4()))
        energy_wh = {}
        capex_breakdown = {}
        opex_breakdown = {}
        tot_install_cost_euro = 0.0
        tot_invest_cost_euro = 0.0
        tot_variable_opex_cost_euro = 0.0
        tot_fixed_opex_cost_euro = 0.0

        for _key, asset in self.esdl_assets.items():
            is_asset_opex_included = False
            if (
                asset.asset_type == "HeatProducer"
                or asset.asset_type == "GenericProducer"
                or asset.asset_type == "ResidualHeatSource"
                or asset.asset_type == "GeothermalSource"
                or asset.asset_type == "ResidualHeatSource"
                or asset.asset_type == "GasHeater"
            ):
                is_asset_opex_included = True

            asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
            placed = np.round(results[asset_placement_var][0]) >= 1.0
            if placed:
                try:
                    capex_breakdown[asset.asset_type] += (
                        results[f"{asset.name}__installation_cost"][0]
                        + results[f"{asset.name}__investment_cost"][0]
                    )
                    tot_install_cost_euro += results[f"{asset.name}__installation_cost"][0]
                    tot_invest_cost_euro += results[f"{asset.name}__investment_cost"][0]

                    if is_asset_opex_included:
                        tot_variable_opex_cost_euro += results[
                            f"{asset.name}__variable_operational_cost"
                        ][0]
                        tot_fixed_opex_cost_euro += results[
                            f"{asset.name}__fixed_operational_cost"
                        ][0]
                        opex_breakdown[asset.asset_type] += (
                            tot_variable_opex_cost_euro + tot_fixed_opex_cost_euro
                        )

                except KeyError:
                    try:
                        capex_breakdown[asset.asset_type] = (
                            results[f"{asset.name}__installation_cost"][0]
                            + results[f"{asset.name}__investment_cost"][0]
                        )
                        tot_install_cost_euro += results[f"{asset.name}__installation_cost"][0]
                        tot_invest_cost_euro += results[f"{asset.name}__investment_cost"][0]

                        if is_asset_opex_included:
                            tot_variable_opex_cost_euro += results[
                                f"{asset.name}__variable_operational_cost"
                            ][0]
                            tot_fixed_opex_cost_euro += results[
                                f"{asset.name}__fixed_operational_cost"
                            ][0]
                            opex_breakdown[asset.asset_type] = (
                                tot_variable_opex_cost_euro + tot_fixed_opex_cost_euro
                            )
                    except KeyError:
                        # Do not add any costs. Items like joint
                        pass

                if is_asset_opex_included:
                    energy_wh[asset.name] = np.sum(
                        results[f"{asset.name}.Heat_source"][1:]
                        * (self.times()[1:] - self.times()[0:-1])
                        / 3600
                    )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="High level cost breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(
                            label="CAPEX",
                            value=tot_install_cost_euro + tot_invest_cost_euro,
                        ),
                        esdl.StringItem(
                            label="OPEX",
                            value=tot_variable_opex_cost_euro + tot_fixed_opex_cost_euro,
                        ),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Overall cost breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label="Installation", value=tot_install_cost_euro),
                        esdl.StringItem(label="Investment", value=tot_invest_cost_euro),
                        esdl.StringItem(label="Variable OPEX", value=tot_variable_opex_cost_euro),
                        esdl.StringItem(label="Fixed OPEX", value=tot_fixed_opex_cost_euro),
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="CAPEX breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in capex_breakdown.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="OPEX breakdown [EUR]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value)
                        for key, value in opex_breakdown.items()
                    ]
                ),
                quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                    physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                ),
            )
        )

        kpis_top_level.kpi.append(
            esdl.DistributionKPI(
                name="Energy production [Wh]",
                distribution=esdl.StringLabelDistribution(
                    stringItem=[
                        esdl.StringItem(label=key, value=value) for key, value in energy_wh.items()
                    ]
                ),
            )
        )
        energy_system.instance[0].area.KPIs = kpis_top_level
        # ------------------------------------------------------------------------------------------
        # Cost breakdowns per polygon areas (can consist of several assets of differents types)
        # Notes:
        # - OPEX KPIs are taken into account for energy sources only.
        # - We assume that all energy produced outside of the the subarea comes in via a heat
        #   exchanger that is part of the subarea.
        # TODO: Investigate if no cost in the ESDL then this breaks ESDL visibility
        for subarea in energy_system.instance[0].area.area:
            area_investment_cost = 0.0
            area_installation_cost = 0.0
            area_variable_opex_cost = 0.0
            area_fixed_opex_cost = 0.0

            is_at_least_asset_opex_included = False
            kpis = esdl.KPIs(id=str(uuid.uuid4()))
            # Here we make a breakdown of the produced energy in the subarea. Where we assume that
            # all energy produced outside of the the subarea comes in via a heat exchanger that is
            # part of the subarea.
            energy_breakdown = {}
            for asset in subarea.asset:
                asset_name = asset.name
                asset_type = self.get_asset_from_asset_name(asset_name).asset_type

                asset_placement_var = self._asset_aggregation_count_var_map[asset.name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0

                if placed:
                    is_asset_opex_included = False
                    if (
                        asset_type == "HeatProducer"
                        or asset_type == "GenericProducer"
                        or asset_type == "ResidualHeatSource"
                        or asset_type == "GeothermalSource"
                        or asset_type == "ResidualHeatSource"
                        or asset_type == "GasHeater"
                        or asset_type == "HeatStorage"
                    ):
                        is_asset_opex_included = True
                        is_at_least_asset_opex_included = True

                    if asset_type == "Joint":
                        continue
                    try:
                        energy_breakdown[asset_type] += np.sum(results[f"{asset_name}.Heat_source"])
                    except KeyError:
                        try:
                            energy_breakdown[asset_type] = np.sum(
                                results[f"{asset_name}.Heat_source"]
                            )
                        except KeyError:
                            try:
                                energy_breakdown[asset_type] += np.sum(
                                    results[f"{asset_name}.Secondary_heat"]
                                )
                            except KeyError:
                                try:
                                    energy_breakdown[asset_type] = np.sum(
                                        results[f"{asset_name}.Secondary_heat"]
                                    )
                                except KeyError:
                                    pass

                    # Create KPIs by using applicable costs for the specific asset
                    area_investment_cost += results[self._asset_investment_cost_map[asset_name]][0]
                    area_installation_cost += results[
                        self._asset_installation_cost_map[asset_name]
                    ][0]
                    if is_asset_opex_included:
                        area_variable_opex_cost += results[
                            self._asset_variable_operational_cost_map[asset_name]
                        ][0]
                        area_fixed_opex_cost += results[
                            self._asset_fixed_operational_cost_map[asset_name]
                        ][0]

            # Here we add KPIs to the polygon area which allows to visualize them by hoovering over
            # it with the mouse
            kpis.kpi.append(
                esdl.DoubleKPI(
                    value=area_investment_cost / 1.0e6,
                    name="Investment",
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                        unit=esdl.UnitEnum.EURO,
                        multiplier=esdl.MultiplierEnum.MEGA,
                    ),
                )
            )
            kpis.kpi.append(
                esdl.DoubleKPI(
                    value=area_installation_cost / 1.0e6,
                    name="Installation",
                    quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                        physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                        unit=esdl.UnitEnum.EURO,
                        multiplier=esdl.MultiplierEnum.MEGA,
                    ),
                )
            )

            if is_at_least_asset_opex_included:
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_variable_opex_cost / 1.0e6,
                        name="Variable OPEX",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )
                kpis.kpi.append(
                    esdl.DoubleKPI(
                        value=area_fixed_opex_cost / 1.0e6,
                        name="Fixed OPEX",
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST,
                            unit=esdl.UnitEnum.EURO,
                            multiplier=esdl.MultiplierEnum.MEGA,
                        ),
                    )
                )

            # Top level KPIs: Cost breakdown in a polygon area (for all assest grouped together)
            kpi_name = f"{subarea.name}: Asset cost breakdown [EUR]"
            if is_asset_opex_included:
                kpis_top_level.kpi.append(
                    esdl.DistributionKPI(
                        name=kpi_name,
                        distribution=esdl.StringLabelDistribution(
                            stringItem=[
                                esdl.StringItem(label="Installation", value=area_installation_cost),
                                esdl.StringItem(label="Investment", value=area_investment_cost),
                                esdl.StringItem(
                                    label="Variable OPEX", value=area_variable_opex_cost
                                ),
                                esdl.StringItem(label="Fixed OPEX", value=area_fixed_opex_cost),
                            ]
                        ),
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                        ),
                    )
                )
            else:
                kpis_top_level.kpi.append(
                    esdl.DistributionKPI(
                        name=kpi_name,
                        distribution=esdl.StringLabelDistribution(
                            stringItem=[
                                esdl.StringItem(label="Installation", value=area_installation_cost),
                                esdl.StringItem(label="Investment", value=area_investment_cost),
                            ]
                        ),
                        quantityAndUnit=esdl.esdl.QuantityAndUnitType(
                            physicalQuantity=esdl.PhysicalQuantityEnum.COST, unit=esdl.UnitEnum.EURO
                        ),
                    )
                )

            # Here we add a distribution KPI to the subarea to which gives a piechart
            # !!!!!!!!!!!!!!! This will only work if the source is in the area?
            # TODO: Still to be resolved since piecharts are still a work in progress in mapeditor
            # kpis.kpi.append(
            #     esdl.DistributionKPI(
            #         name="Energy breakdown ?",
            #         distribution=esdl.StringLabelDistribution(
            #             stringItem=[
            #                 esdl.StringItem(label=key, value=value) for key,
            #                 value in energy_breakdown.items()
            #             ]
            #         )
            #     )
            # )
            subarea.KPIs = kpis
        # end KPIs
        # ------------------------------------------------------------------------------------------
        # Placement
        for _, attributes in self.esdl_assets.items():
            name = attributes.name
            if name in [
                *self.heat_network_components.get("source", []),
                *self.heat_network_components.get("ates", []),
                *self.heat_network_components.get("buffer", []),
            ]:
                asset = _name_to_asset(name)
                asset_placement_var = self._asset_aggregation_count_var_map[name]
                placed = np.round(results[asset_placement_var][0]) >= 1.0
                max_size = results[self._asset_max_size_map[name]][0]

                if asset in self.heat_network_components.get("buffer", []):
                    asset.capacity = max_size
                    asset.volume = max_size / (
                        parameters[f"{name}.cp"]
                        * parameters[f"{name}.rho"]
                        * parameters[f"{name}.dT"]
                    )
                else:
                    asset.power = max_size
                if not placed:
                    asset.delete(recursive=True)
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED

        if db_profiles:
            for name in [
                *self.heat_network_components.get("source", []),
                *self.heat_network_components.get("demand", []),
                *self.heat_network_components.get("pipe", []),
                *self.heat_network_components.get("buffer", []),
                *self.heat_network_components.get("ates", []),
                *self.heat_network_components.get("heat_exchanger", []),
                *self.heat_network_components.get("heat_pump", []),
            ]:
                asset = _name_to_asset(name)
                try:
                    # For all components dealing with one hydraulic system
                    for variable in ["Heat_flow", "HeatIn.Q", "HeatIn.H"]:
                        profile = esdl.InfluxDBProfile(
                            database="GROW results",
                            measurement=name,
                            field=variable,
                            port=443,
                            host="localhost",
                        )
                        asset.port[1].profile.append(profile)
                except Exception:
                    # For all components dealing with two hydraulic system
                    for variable in [
                        "Heat_flow",
                        "Primary.HeatIn.Q",
                        "Primary.HeatIn.H",
                        "Secondary.HeatIn.Q",
                        "Secondary.HeatIn.H",
                    ]:
                        profile = esdl.InfluxDBProfile(
                            database="GROW results",
                            measurement=name,
                            field=variable,
                            port=443,
                            host="localhost",
                        )
                        asset.port[1].profile.append(profile)

        # Pipes:
        edr_pipe_properties_to_copy = ["innerDiameter", "outerDiameter", "diameter", "material"]

        esh_edr = EnergySystemHandler()

        for pipe in self.hot_pipes:
            pipe_classes = self.pipe_classes(pipe)
            # When a pipe has not been optimized, enforce pipe to be shown in the simulator
            # ESDL.
            if not pipe_classes:
                if optimizer_sim:
                    continue
                else:
                    asset.state = esdl.AssetStateEnum.ENABLED

            if optimizer_sim:
                pipe_class = self.get_optimized_pipe_class(pipe)
            cold_pipe = self.hot_to_cold_pipe(pipe)

            if parameters[f"{pipe}.diameter"] != 0.0 or any(np.abs(results[f"{pipe}.Q"]) > 1.0e-9):
                # if not isinstance(pipe_class, EDRPipeClass):
                #     assert pipe_class.name == f"{pipe}_orig"
                #     continue
                # print(results[f"{pipe}.Q"])
                # print(pipe + " has pipeclass: " + pipe_class.name )
                # print(pipe + f" has diameter: " + pipe_class.name)

                if optimizer_sim:
                    assert isinstance(pipe_class, EDRPipeClass)

                    asset_edr = esh_edr.load_from_string(pipe_class.xml_string)

                for p in [pipe, cold_pipe]:
                    asset = _name_to_asset(p)
                    asset.state = esdl.AssetStateEnum.ENABLED
                    if optimizer_sim:
                        for prop in edr_pipe_properties_to_copy:
                            setattr(asset, prop, getattr(asset_edr, prop))
            else:
                for p in [pipe, cold_pipe]:
                    asset = _name_to_asset(p)
                    asset.delete(recursive=True)

        self.__optimized_energy_system_handler = esh
        self.optimized_esdl_string = esh.to_string()

        if self.esdl_string is None:
            if optimizer_sim:
                filename = run_info.esdl_file.with_name(
                    f"{run_info.esdl_file.stem}_GrowOptimized.esdl"
                )
            else:
                filename = run_info.esdl_file.with_name(
                    f"{run_info.esdl_file.stem}_Simulation.esdl"
                )
            esh.save(str(filename))



import math

import numpy as np

from .constants import GRAVITATIONAL_CONSTANT


def _kinematic_viscosity(temperature):
    """
    The kinematic viscosity barely changes with pressure. The below polynomial
    fit has been deduced using the `iapws` package, which implements the IAPWS
    standard. The valid range of temperatures is 20 - 130 C.

    Below the snippet of code used to get the polynomial coefficients
    (possibly useful when refitting):

    .. code-block:: python

        from iapws import IAPWS95

        import numpy as np

        res = []
        for t in np.linspace(20, 130, 1000):
            res.append((t, IAPWS95(T=273.15 + t, P=0.5).nu))
        print(np.polyfit(*zip(*res), 4))
    """

    if temperature < 20 or temperature > 130:
        raise Exception(
            "Temperature should be in the range 20 - 130 °C.\n"
            "Note that we use Celcius as the unit, not Kelvin."
        )

    return np.polyval(
        [7.53943453e-15, -3.01485854e-12, 4.75924986e-10, -3.79135487e-08, 1.58737429e-06],
        temperature,
    )


def _colebrook_white(reynolds, relative_roughness, friction_factor=0.015):
    for _ in range(1000):
        friction_factor_old = friction_factor

        reynolds_star = (
            1 / math.sqrt(8.0) * reynolds * math.sqrt(friction_factor) * relative_roughness
        )
        friction_factor = (
            1.0
            / (
                -2.0
                * math.log10(
                    2.51 / reynolds / math.sqrt(friction_factor) * (1 + reynolds_star / 3.3)
                )
            )
            ** 2
        )

        if (
            abs(friction_factor - friction_factor_old) / max(friction_factor, friction_factor_old)
            < 1e-6
        ):
            return friction_factor
    else:
        raise Exception("Colebrook-White did not converge")


def friction_factor(velocity, diameter, wall_roughness, temperature):
    """
    Darcy-weisbach friction factor calculation from both laminar and turbulent
    flow.
    """

    kinematic_viscosity = _kinematic_viscosity(temperature)
    reynolds = velocity * diameter / kinematic_viscosity

    assert velocity >= 0

    if velocity == 0.0 or diameter == 0.0:
        return 0.0
    elif reynolds <= 2000.0:
        friction_factor = 64.0 / reynolds
    elif reynolds >= 4000.0:
        friction_factor = _colebrook_white(reynolds, wall_roughness / diameter)
    else:
        fac_turb = _colebrook_white(4000.0, wall_roughness / diameter)
        fac_laminar = 64.0 / 2000.0
        w = (reynolds - 2000.0) / 2000.0
        friction_factor = w * fac_turb + (1 - w) * fac_laminar

    return friction_factor


def head_loss(velocity, diameter, length, wall_roughness, temperature):
    """
    Head loss for a circular pipe of given length.
    """

    f = friction_factor(velocity, diameter, wall_roughness, temperature)

    return length * f / (2 * GRAVITATIONAL_CONSTANT) * velocity**2 / diameter


def get_linear_pipe_dh_vs_q_fit(
    diameter, length, wall_roughness, temperature, n_lines=10, v_max=2.5
):
    area = math.pi * diameter**2 / 4

    v_points = np.linspace(0.0, v_max, n_lines + 1)
    q_points = v_points * area

    h_points = np.array(
        [head_loss(v, diameter, length, wall_roughness, temperature) for v in v_points]
    )

    a = np.diff(h_points) / np.diff(q_points)
    b = h_points[1:] - a * q_points[1:]

    return a, b


def get_linear_pipe_power_hydraulic_vs_q_fit(
    rho, diameter, length, wall_roughness, temperature, n_lines=10, v_max=2.5
):
    """
    power_hydraulic = b + (a * Q)
    """
    area = math.pi * diameter**2 / 4.0

    v_points = np.linspace(0.0, v_max, n_lines + 1)
    q_points = v_points * area
    power_hydraulic_points = np.array(
        [
            rho
            * GRAVITATIONAL_CONSTANT
            * abs(head_loss(v, diameter, length, wall_roughness, temperature))
            * v
            * area
            for v in v_points
        ]
    )

    a = np.diff(power_hydraulic_points) / np.diff(q_points)  # calc gradients for n_line segments
    b = power_hydraulic_points[1:] - a * q_points[1:]

    return a, b



from pathlib import Path
from unittest import TestCase

import numpy as np

import pandas as pd

from rtctools.util import run_optimization_problem

from rtctools_heat_network.head_loss_mixin import HeadLossOption


class TestHydraulicPower(TestCase):
    def test_hydraulic_power(self):
        import models.pipe_test.src.run_hydraulic_power as run_hydraulic_power
        from models.pipe_test.src.run_hydraulic_power import (
            HeatProblem,
        )

        # Settings
        base_folder = Path(run_hydraulic_power.__file__).resolve().parent.parent
        run_hydraulic_power.comp_vars_vals = {
            "pipe_length": [25000.0],  # [m]
        }
        run_hydraulic_power.comp_vars_init = {
            "pipe_length": 0.0,  # [m]
            "heat_demand": [3.95 * 10**6, 3.95 * 10**6],  # [W]
            "pipe_DN_MILP": 300,  # [mm]
        }
        standard_columns_specified = [
            "Pipe1_supply_dPress",
            "Pipe1_return_dPress",
            "Pipe1_supply_Q",
            "Pipe1_return_Q",
            "Pipe1_supply_mass_flow",
            "Pipe1_return_mass_flow",
            "Pipe1_supply_flow_vel",
            "Pipe1_return_flow_vel",
            "Pipe1_supply_dT",
            "Pipe1_return_dT",
            "Heat_source",
            "Heat_demand",
            "Heat_loss",
            "pipe_length",
        ]

        # Initialize variables
        run_hydraulic_power.ThermalDemand = run_hydraulic_power.comp_vars_init["heat_demand"]
        run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_init[
            "pipe_length"
        ]
        run_hydraulic_power.manual_set_pipe_DN_diam_MILP = run_hydraulic_power.comp_vars_init[
            "pipe_DN_MILP"
        ]
        # ----------------------------------------------------------------------------------------
        # 3 MILP simulations with the only difference being the linear head loss setting:
        # - LINEARIZED_DW (1 line segment)
        # - LINEAR
        # - LINEARIZED_DW (default line segments = 5)
        # ----------------------------------------------------------------------------------------
        # Run MILP with LINEARIZED_DW head loss setting and 1 line segement
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEARIZED_DW
        run_hydraulic_power.n_linearization_lines_setting = 1

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_optimization_problem(HeatProblem, base_folder=base_folder)

        hydraulic_power_post_process_dw_1 = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][0] * abs(
            run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0]
        )
        hydraulic_power_dw_1 = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (hydraulic_power_dw_1) of the 3rd order function should
        # overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_dw_1,
            hydraulic_power_dw_1,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Compare hydraulic power, for an one hour timeseries with a specific demand, to a hard
        # coded value which originates from runnning MILP without big_m method being implemented,
        # during the comparison of MILP and a high-fidelity code
        np.testing.assert_allclose(
            104829.66021214866,
            hydraulic_power_dw_1,
        )
        np.testing.assert_allclose(
            run_hydraulic_power.df_MILP["Pipe1_return_Hydraulic_power"][0],
            run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0],
            rtol=1e-2,
        )
        # ----------------------------------------------------------------------------------------
        # Rerun MILP with LINEAR head loss setting
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)  # empty df
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEAR

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_optimization_problem(HeatProblem, base_folder=base_folder)

        hydraulic_power_post_process_linear = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][
            0
        ] * abs(run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0])
        hydraulic_power_linear = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (1 line segment) of the 3rd order function should
        # overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_linear,
            hydraulic_power_linear,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Hydraulic hydraulic =  delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The predicted hydraulic power should be the same when the delta pressure is approximated
        # by a linear segment, via 2 different head loss setting options. Head loss setting =
        # HeadLossOption.LINEAR and HeadLossOption.LINEARIZED_DW (with 1 linear segment)
        np.testing.assert_allclose(
            hydraulic_power_post_process_linear,
            hydraulic_power_post_process_dw_1,
            rtol=1e-7,
            err_msg="Values should be the same",
        )
        # Hydraulic hydraulic =  delta pressure * Q = f(Q^3)
        # The predicted hydraulic power should be the same if it is approximated by a linear segment
        # , via 2 different head loss setting options. Head loss setting =
        # HeadLossOption.LINEAR and HeadLossOption.LINEARIZED_DW (with 1 linear segment)
        np.testing.assert_allclose(
            hydraulic_power_linear, hydraulic_power_dw_1, err_msg="Values should be the same"
        )
        # ----------------------------------------------------------------------------------------
        # Rerun MILP with DW head loss setting, and default line segments
        run_hydraulic_power.df_MILP = pd.DataFrame(columns=standard_columns_specified)  # empty df
        run_hydraulic_power.head_loss_setting = HeadLossOption.LINEARIZED_DW
        run_hydraulic_power.n_linearization_lines_setting = 5

        for val in range(0, len(run_hydraulic_power.comp_vars_vals["pipe_length"])):
            run_hydraulic_power.manual_set_pipe_length = run_hydraulic_power.comp_vars_vals[
                "pipe_length"
            ][val]
            run_optimization_problem(HeatProblem, base_folder=base_folder)

        hydraulic_power_post_process_dw = run_hydraulic_power.df_MILP["Pipe1_supply_Q"][0] * abs(
            run_hydraulic_power.df_MILP["Pipe1_supply_dPress"][0]
        )
        hydraulic_power_dw = run_hydraulic_power.df_MILP["Pipe1_supply_Hydraulic_power"][0]
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation (default number of line segments) of the 3rd order function
        # should overestimate the hydraulic power when compared to the product of Q and the linear
        # approximation (default number of line segments) of 2nd order function (delta pressure).
        np.testing.assert_array_less(
            hydraulic_power_post_process_dw,
            hydraulic_power_dw,
            "Post process hydraulic power must be < hydraulic_power",
        )
        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The approximation of the 3rd order function via 5 line segments (dafault value) should be
        # better compared to 1 line segment approximation thereof. The latter will result in an
        # overstimated prediction
        np.testing.assert_array_less(
            hydraulic_power_dw,
            hydraulic_power_dw_1,
            "5 line segments predicted hydraulic power > hydraulic_power with 1 line segment",
        )
        # Compare hydraulic power, for an one hour timeseries with a specific demand, to a hard
        # coded value which originates from runnning MILP without big_m method being implemented,
        # during the comparison of MILP and a high-fidelity code
        np.testing.assert_allclose(
            4367.240507173596,
            hydraulic_power_dw,
        )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestHydraulicPower()
    a.test_hydraulic_power()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



import logging

import esdl
from esdl import InPort


from rtctools_heat_network.pycml import Model as _Model

logger = logging.getLogger("rtctools_heat_network")


RETRY_LOOP_LIMIT = 100


class _RetryLaterException(Exception):
    pass


class _SkipAssetException(Exception):
    pass


class _ESDLModelBase(_Model):
    def _esdl_convert(self, converter, assets, prefix):
        # Sometimes we need information of one component in order to convert
        # another. For example, the nominal discharg of a pipe is used to set
        # the nominal discharge of its connected components.
        retry_assets = list(assets.values())
        skip_assets = list()

        for _ in range(RETRY_LOOP_LIMIT):
            current_assets = retry_assets
            retry_assets = []

            for asset in current_assets:
                try:
                    pycml_type, modifiers = converter.convert(asset)
                except _SkipAssetException:
                    skip_assets.append(asset)
                    continue
                except _RetryLaterException:
                    retry_assets.append(asset)
                    continue

                self.add_variable(pycml_type, asset.name, **modifiers)

            if not retry_assets:
                break
        else:
            raise Exception("Parsing of assets exceeded maximum iteration limit.")

        in_suf = f"{prefix}In"
        out_suf = f"{prefix}Out"
        node_suf = f"{prefix}Conn"
        elec_in_suf = "ElectricityIn"
        elec_out_suf = "ElectricityOut"
        elec_node_suf = "ElectricityConn"
        gas_in_suf = "GasIn"
        gas_out_suf = "GasOut"
        gas_node_suf = "GasConn"

        skip_asset_ids = {a.id for a in skip_assets}
        pipe_assets = [
            a
            for a in assets.values()
            if (
                a.asset_type == "Pipe"
                and a.id not in skip_asset_ids
                and isinstance(a.in_ports[0].carrier, esdl.HeatCommodity)
            )
        ]
        node_assets = [
            a
            for a in assets.values()
            if (
                (a.asset_type == "Joint")
                and a.id not in skip_asset_ids
                and (
                    (isinstance(a.in_ports[0].carrier, esdl.HeatCommodity))
                    or isinstance(a.out_ports[0].carrier, esdl.HeatCommodity)
                )
            )
        ]
        gas_node_assets = [
            a
            for a in assets.values()
            if (
                (a.asset_type == "Joint")
                and a.id not in skip_asset_ids
                and (
                    (isinstance(a.in_ports[0].carrier, esdl.GasCommodity))
                    or isinstance(a.out_ports[0].carrier, esdl.GasCommodity)
                )
            )
        ]
        bus_assets = [
            a for a in assets.values() if ((a.asset_type == "Bus") and a.id not in skip_asset_ids)
        ]
        non_node_assets = [
            a
            for a in assets.values()
            if (a.asset_type != "Joint" and a.asset_type != "Bus") and a.id not in skip_asset_ids
        ]

        # Here we check that every pipe and node has the correct coupled carrier for their _ret
        # asset.
        for asset in [*pipe_assets, *node_assets]:
            asset_carrier = asset.global_properties["carriers"][asset.in_ports[0].carrier.id]
            couple_asset_carrier = next(
                x.global_properties["carriers"][x.in_ports[0].carrier.id]
                for x in [*pipe_assets, *node_assets]
                if (
                    x.name.replace("_ret", "") == asset.name.replace("_ret", "")
                    and x.name != asset.name
                )
            )
            if asset_carrier["name"] != couple_asset_carrier["name"]:
                raise Exception(
                    f"{asset.name} and {asset.name}_ret do not have the matching "
                    f"carriers specified"
                )

        # First we map all port ids to their respective PyCML ports. We only
        # do this for non-nodes, as for nodes we don't quite know what port
        # index a connection has to use yet.
        port_map = {}

        for asset in non_node_assets:
            component = getattr(self, asset.name)
            # We assume that every component has 2 ports. Essentially meaning that we are dealing
            # with a single commodity for a component. Exceptions, assets that deal with multiple
            # have to be specifically specified what port configuration is expected in the model.
            if (
                asset.asset_type == "GenericConversion"
                or asset.asset_type == "HeatExchange"
                or asset.asset_type == "HeatPump"
            ):
                if prefix != "Heat":
                    raise Exception(
                        "Hydraulically decoulpled systems are not yet supported for nonlinear (QTH)"
                        "optimization"
                    )
                # check for expected number of ports
                if len(asset.in_ports) == 2 and len(asset.out_ports) == 2:
                    for p in [*asset.in_ports, *asset.out_ports]:
                        if isinstance(p.carrier, esdl.HeatCommodity):
                            if isinstance(p, InPort):
                                if "_ret" in p.carrier.name:
                                    port_map[p.id] = getattr(component.Secondary, in_suf)
                                else:
                                    port_map[p.id] = getattr(component.Primary, in_suf)
                            else:  # OutPort
                                if "_ret" in p.carrier.name:
                                    port_map[p.id] = getattr(component.Primary, out_suf)
                                else:
                                    port_map[p.id] = getattr(component.Secondary, out_suf)
                        else:
                            raise Exception(
                                f"{asset.name} has does not have 2 Heat in_ports and 2 Heat "
                                f"out_ports "
                            )
                elif len(asset.in_ports) == 3 and len(asset.out_ports) == 2:
                    p_heat = 0
                    p_elec = 0
                    for p in [*asset.in_ports, *asset.out_ports]:
                        if isinstance(p.carrier, esdl.HeatCommodity) and p_heat <= 3:
                            if isinstance(p, InPort):
                                if "_ret" in p.carrier.name:
                                    port_map[p.id] = getattr(component.Secondary, in_suf)
                                else:
                                    port_map[p.id] = getattr(component.Primary, in_suf)
                            else:  # OutPort
                                if "_ret" in p.carrier.name:
                                    port_map[p.id] = getattr(component.Primary, out_suf)
                                else:
                                    port_map[p.id] = getattr(component.Secondary, out_suf)
                            p_heat += 1
                        elif isinstance(p.carrier, esdl.ElectricityCommodity) and p_elec == 0:
                            port_map[p.id] = getattr(component, elec_in_suf)
                            p_elec += 1
                        else:
                            raise Exception(
                                f"{asset.name} has total of 5 ports, but no proper split between "
                                f"heat(4) and electricity (1) ports"
                            )

                else:
                    raise Exception(
                        f"{asset.name} has does not have 2 or 3 in_ports and 2 " f"out_ports "
                    )
            elif (
                asset.in_ports is None
                and isinstance(asset.out_ports[0].carrier, esdl.ElectricityCommodity)
                and len(asset.out_ports) == 1
            ):
                port_map[asset.out_ports[0].id] = getattr(component, elec_out_suf)
            elif (
                asset.in_ports is None
                and isinstance(asset.out_ports[0].carrier, esdl.GasCommodity)
                and len(asset.out_ports) == 1
            ):
                port_map[asset.out_ports[0].id] = getattr(component, gas_out_suf)
            elif (
                len(asset.in_ports) == 1
                and isinstance(asset.in_ports[0].carrier, esdl.ElectricityCommodity)
                and asset.out_ports is None
            ):
                port_map[asset.in_ports[0].id] = getattr(component, elec_in_suf)
            elif (
                len(asset.in_ports) == 1
                and isinstance(asset.in_ports[0].carrier, esdl.GasCommodity)
                and asset.out_ports is None
            ):
                port_map[asset.in_ports[0].id] = getattr(component, gas_in_suf)
            elif (
                len(asset.in_ports) == 1
                and isinstance(asset.in_ports[0].carrier, esdl.HeatCommodity)
                and len(asset.out_ports) == 1
            ):
                port_map[asset.in_ports[0].id] = getattr(component, in_suf)
                port_map[asset.out_ports[0].id] = getattr(component, out_suf)
            elif (
                len(asset.in_ports) == 1
                and isinstance(asset.in_ports[0].carrier, esdl.ElectricityCommodity)
                and len(asset.out_ports) == 1
            ):
                port_map[asset.in_ports[0].id] = getattr(component, elec_in_suf)
                port_map[asset.out_ports[0].id] = getattr(component, elec_out_suf)
            elif (
                len(asset.in_ports) == 1
                and isinstance(asset.in_ports[0].carrier, esdl.GasCommodity)
                and len(asset.out_ports) == 1
            ):
                port_map[asset.in_ports[0].id] = getattr(component, gas_in_suf)
                port_map[asset.out_ports[0].id] = getattr(component, gas_out_suf)
            else:
                raise Exception(f"Unsupported ports for asset type {asset.name}.")

        # Nodes are special in that their in/out ports can have multiple
        # connections. This means we have some bookkeeping to do per node. We
        # therefore do the nodes first, and do all remaining connections
        # after.
        connections = set()

        for asset in [*node_assets, *bus_assets, *gas_node_assets]:
            component = getattr(self, asset.name)

            i = 1
            if len(asset.in_ports) != 1 or len(asset.out_ports) != 1:
                Exception(
                    f"{asset.name} has !=1 in or out ports, please only use one, "
                    f"multiple connections to a single joint port are allowed"
                )
            for port in (asset.in_ports[0], asset.out_ports[0]):
                for connected_to in port.connectedTo.items:
                    conn = (port.id, connected_to.id)
                    if conn in connections or tuple(reversed(conn)) in connections:
                        continue
                    if isinstance(port.carrier, esdl.HeatCommodity):
                        self.connect(getattr(component, node_suf)[i], port_map[connected_to.id])
                        connections.add(conn)
                        i += 1
                    elif isinstance(port.carrier, esdl.ElectricityCommodity):
                        self.connect(
                            getattr(component, elec_node_suf)[i], port_map[connected_to.id]
                        )
                        connections.add(conn)
                        i += 1
                    elif isinstance(port.carrier, esdl.GasCommodity):
                        self.connect(getattr(component, gas_node_suf)[i], port_map[connected_to.id])
                        connections.add(conn)
                        i += 1
                    else:
                        logger.error(
                            f"asset {asset.name} has an unsupported carrier type for a node"
                        )

        skip_port_ids = set()
        for a in skip_assets:
            if a.in_ports is not None:
                for port in a.in_ports:
                    skip_port_ids.add(port.id)
            if a.out_ports is not None:
                for port in a.out_ports:
                    skip_port_ids.add(port.id)

        # All non-Joints/nodes
        for asset in non_node_assets:
            ports = []
            if asset.in_ports is not None:
                ports.extend(asset.in_ports)
            if asset.out_ports is not None:
                ports.extend(asset.out_ports)
            assert len(ports) > 0
            for port in ports:
                connected_ports = [p for p in port.connectedTo.items if p.id not in skip_port_ids]
                if len(connected_ports) != 1:
                    logger.warning(
                        f"{asset.asset_type} '{asset.name}' has multiple connections"
                        f" to a single port. "
                    )

                assert len(connected_ports) == 1

                for connected_to in connected_ports:
                    conn = (port.id, connected_to.id)
                    if conn in connections or tuple(reversed(conn)) in connections:
                        continue

                    self.connect(port_map[port.id], port_map[connected_to.id])
                    connections.add(conn)



from typing import Dict

import numpy as np

from rtctools.optimization.timeseries import Timeseries

from .base_component_type_mixin import BaseComponentTypeMixin
from .heat_network_common import PipeFlowDirection


class BoundsToPipeFlowDirectionsMixin(BaseComponentTypeMixin):
    def pre(self):
        super().pre()

        bounds = self.bounds()
        components = self.heat_network_components
        pipes = components["pipe"]

        # Determine implied pipe directions from model bounds (that are
        # already available at this time)
        self.__implied_directions = [{} for e in range(self.ensemble_size)]

        for e in range(self.ensemble_size):
            for p in pipes:
                lb, ub = bounds[f"{p}.Q"]

                if not isinstance(lb, float) or not isinstance(ub, float):
                    raise ValueError(
                        "`BoundsToPipeFlowDirectionsMixin` only works for scalar bounds"
                    )

                if lb == ub and lb == 0.0:
                    # Pipe is disabled
                    self.__implied_directions[e][p] = PipeFlowDirection.DISABLED
                elif lb >= 0.0:
                    self.__implied_directions[e][p] = PipeFlowDirection.POSITIVE
                elif ub <= 0.0:
                    self.__implied_directions[e][p] = PipeFlowDirection.NEGATIVE

    def constant_inputs(self, ensemble_member):
        inputs = super().constant_inputs(ensemble_member)
        for p, d in self.__implied_directions[ensemble_member].items():
            k = self.heat_network_flow_directions[p]
            inputs[k] = Timeseries([-np.inf, np.inf], [d, d])
        return inputs

    @property
    def heat_network_flow_directions(self) -> Dict[str, str]:
        pipes = self.heat_network_components["pipe"]
        return {p: f"{p}__implied_direction" for p in pipes}



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class ControlValve(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "control_valve"

        self.add_variable(Variable, "dH")

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



import math
from dataclasses import dataclass
from typing import Tuple


@dataclass(frozen=True)
class PipeClass:
    name: str
    inner_diameter: float
    maximum_velocity: float
    u_values: Tuple[float, float]
    investment_costs: float

    @property
    def maximum_discharge(self):
        return self.area * self.maximum_velocity

    @property
    def area(self):
        return 0.25 * math.pi * self.inner_diameter**2



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class Pump(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(Q=dict(min=0.0)),
                modifiers,
            ),
        )

        self.component_type = "pump"

        self.add_variable(Variable, "dH", min=0.0)

        self.add_equation(self.dH - (self.QTHOut.H - self.QTHIn.H))

        self.add_equation(self.QTHIn.T - self.QTHOut.T)



import logging
from typing import Dict

from pymoca.backends.casadi.alias_relation import AliasRelation

from .base_component_type_mixin import BaseComponentTypeMixin
from .heat_network_common import NodeConnectionDirection
from .topology import Topology

logger = logging.getLogger("rtctools_heat_network")


class ModelicaComponentTypeMixin(BaseComponentTypeMixin):
    def pre(self):
        components = self.heat_network_components
        nodes = components.get("node", [])
        busses = components.get("electricity_node", [])
        gas_nodes = components.get("gas_node", [])
        buffers = components.get("buffer", [])
        atess = components.get("ates", [])
        try:
            pipes = components["pipe"]
            cables = components.get("electricity_cable", [])
            gas_pipes = components.get("gas_pipe", [])
        except KeyError:
            try:
                cables = components["electricity_cable"]
                gas_pipes = components.get("gas_pipe", [])
                pipes = []
            except KeyError:
                try:
                    cables = []
                    gas_pipes = components["gas_pipe"]
                    pipes = []
                except KeyError:
                    logger.error(
                        "A valid network should have at least one pipe/cable, "
                        "assets cannot be connected directly"
                    )

        # Figure out which pipes are connected to which nodes, which pipes
        # are connected in series, and which pipes are connected to which buffers.

        pipes_set = set(pipes)
        cables_set = set(cables)
        gas_pipes_set = set(gas_pipes)
        parameters = [self.parameters(e) for e in range(self.ensemble_size)]
        node_connections = {}
        bus_connections = {}
        gas_node_connections = {}

        # Figure out if we are dealing with a Heat model, or a QTH model
        try:
            if len(pipes):
                _ = self.variable(f"{pipes[0]}.HeatIn.Heat")
                heat_network_model_type = "Heat"
            else:
                heat_network_model_type = "Heat"
        except KeyError:
            heat_network_model_type = "QTH"

        for n in [*nodes, *busses, *gas_nodes]:
            n_connections = [ens_params[f"{n}.n"] for ens_params in parameters]

            if len(set(n_connections)) > 1:
                raise Exception(
                    "Nodes and busses cannot have differing number of connections per "
                    "ensemble member"
                )

            n_connections = n_connections[0]

            # Note that we do this based on temperature, because discharge may
            # be an alias of yet some other further away connected pipe.
            if n in nodes:
                node_connections[n] = connected_pipes = {}
            elif n in busses:
                bus_connections[n] = connected_pipes = {}
            elif n in gas_nodes:
                gas_node_connections[n] = connected_pipes = {}

            for i in range(n_connections):
                if n in nodes:
                    cur_port = f"{n}.{heat_network_model_type}Conn[{i + 1}]"
                    prop = "T" if heat_network_model_type == "QTH" else "Heat"
                    in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                    out_suffix = (
                        ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
                    )
                elif n in busses:
                    cur_port = f"{n}.ElectricityConn[{i + 1}]"
                    prop = "Power"
                    in_suffix = ".ElectricityIn.Power"
                    out_suffix = ".ElectricityOut.Power"
                elif n in gas_nodes:
                    # TODO: Ideally a temporary variable would be created to make the connections
                    #  map that is not passed to the problem
                    cur_port = f"{n}.GasConn[{i + 1}]"
                    prop = "Q_shadow"
                    in_suffix = ".GasIn.Q_shadow"
                    out_suffix = ".GasOut.Q_shadow"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{cur_port}.{prop}")
                    if not x.startswith(n) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {cur_port}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {cur_port}")

                if aliases[0].endswith(out_suffix):
                    pipe_w_orientation = (
                        aliases[0][: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert aliases[0].endswith(in_suffix)
                    pipe_w_orientation = (
                        aliases[0][: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert (
                    pipe_w_orientation[0] in pipes_set
                    or pipe_w_orientation[0] in cables_set
                    or pipe_w_orientation[0] in gas_pipes_set
                )

                connected_pipes[i] = pipe_w_orientation

        # Note that a pipe series can include both hot and cold pipes for
        # QTH models. It is only about figuring out which pipes are
        # related direction-wise.
        # For Heat models, only hot pipes are allowed to be part of pipe
        # series, as the cold part is zero heat by construction.
        if heat_network_model_type == "QTH":
            alias_relation = self.alias_relation
        elif heat_network_model_type == "Heat":
            # There is no proper AliasRelation yet (because there is heat loss in pipes).
            # So we build one, as that is the easiest way to figure out which pipes are
            # connected to each other in series. We do this by making a temporary/shadow
            # discharge (".Q") variable per pipe, as that way we can share the processing
            # logic for determining pipe series with that of QTH models.
            alias_relation = AliasRelation()

            # Look for aliases only in the hot pipes. All cold pipes are zero by convention anyway.
            hot_pipes = self.hot_pipes.copy()

            pipes_map = {f"{pipe}.HeatIn.Heat": pipe for pipe in hot_pipes}
            pipes_map.update({f"{pipe}.HeatOut.Heat": pipe for pipe in hot_pipes})

            for p in hot_pipes:
                for port in ["In", "Out"]:
                    heat_port = f"{p}.Heat{port}.Heat"
                    connected = self.alias_relation.aliases(heat_port).intersection(
                        pipes_map.keys()
                    )
                    connected.remove(heat_port)

                    if connected:
                        other_pipe_port = next(iter(connected))
                        if other_pipe_port.endswith(f".Heat{port}.Heat"):
                            sign_prefix = "-"
                        else:
                            sign_prefix = ""
                        other_pipe = pipes_map[other_pipe_port]
                        if f"{other_pipe}.Q" not in alias_relation.canonical_variables:
                            alias_relation.add(f"{p}.Q", f"{sign_prefix}{other_pipe}.Q")

        canonical_pipe_qs = {p: alias_relation.canonical_signed(f"{p}.Q") for p in pipes}
        # Move sign from canonical to alias
        canonical_pipe_qs = {(p, d): c for p, (c, d) in canonical_pipe_qs.items()}
        # Reverse the dictionary from `Dict[alias, canonical]` to `Dict[canonical, Set[alias]]`
        pipe_sets = {}
        for a, c in canonical_pipe_qs.items():
            pipe_sets.setdefault(c, []).append(a)

        pipe_series_with_orientation = list(pipe_sets.values())

        # Check that all pipes in the series have the same orientation
        pipe_series = []
        for ps in pipe_series_with_orientation:
            if not len({orientation for _, orientation in ps}) == 1:
                raise Exception(f"Pipes in series {ps} do not all have the same orientation")
            pipe_series.append([name for name, _ in ps])

        buffer_connections = {}

        for b in buffers:
            buffer_connections[b] = []

            for k in ["In", "Out"]:
                b_conn = f"{b}.{heat_network_model_type}{k}"
                prop = "T" if heat_network_model_type == "QTH" else "Heat"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{b_conn}.{prop}")
                    if not x.startswith(b) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {b_conn}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {b_conn}")

                in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                out_suffix = ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"
                alias = aliases[0]
                if alias.endswith(out_suffix):
                    pipe_w_orientation = (
                        alias[: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert alias.endswith(in_suffix)
                    pipe_w_orientation = (
                        alias[: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert pipe_w_orientation[0] in pipes_set

                if k == "In":
                    assert self.is_hot_pipe(pipe_w_orientation[0])
                else:
                    assert self.is_cold_pipe(pipe_w_orientation[0])

                buffer_connections[b].append(pipe_w_orientation)

            buffer_connections[b] = tuple(buffer_connections[b])

        ates_connections = {}

        for a in atess:
            ates_connections[a] = []

            for k in ["In", "Out"]:
                a_conn = f"{a}.{heat_network_model_type}{k}"
                prop = "T" if heat_network_model_type == "QTH" else "Heat"
                aliases = [
                    x
                    for x in self.alias_relation.aliases(f"{a_conn}.{prop}")
                    if not x.startswith(a) and x.endswith(f".{prop}")
                ]

                if len(aliases) > 1:
                    raise Exception(f"More than one connection to {a_conn}")
                elif len(aliases) == 0:
                    raise Exception(f"Found no connection to {a_conn}")

                in_suffix = ".QTHIn.T" if heat_network_model_type == "QTH" else ".HeatIn.Heat"
                out_suffix = ".QTHOut.T" if heat_network_model_type == "QTH" else ".HeatOut.Heat"

                if aliases[0].endswith(out_suffix):
                    pipe_w_orientation = (
                        aliases[0][: -len(out_suffix)],
                        NodeConnectionDirection.IN,
                    )
                else:
                    assert aliases[0].endswith(in_suffix)
                    pipe_w_orientation = (
                        aliases[0][: -len(in_suffix)],
                        NodeConnectionDirection.OUT,
                    )

                assert pipe_w_orientation[0] in pipes_set

                if k == "Out":
                    assert self.is_cold_pipe(pipe_w_orientation[0])
                else:
                    assert self.is_hot_pipe(pipe_w_orientation[0])

                ates_connections[a].append(pipe_w_orientation)

            ates_connections[a] = tuple(ates_connections[a])

        self.__topology = Topology(
            node_connections,
            gas_node_connections,
            pipe_series,
            buffer_connections,
            ates_connections,
            bus_connections,
        )

        super().pre()

    @property
    def heat_network_components(self) -> Dict[str, str]:
        try:
            return self.__hn_component_types
        except AttributeError:
            string_parameters = self.string_parameters(0)

            # Find the components in model, detection by string
            # (name.component_type: type)
            component_types = sorted({v for k, v in string_parameters.items()})

            components = {}
            for c in component_types:
                components[c] = sorted(
                    {k.split(".")[0] for k, v in string_parameters.items() if v == c}
                )

            self.__hn_component_types = components

            return components

    @property
    def heat_network_topology(self) -> Topology:
        return self.__topology



from numpy import nan

from .qth_two_port import QTHTwoPort


class _FluidPropertiesComponent(QTHTwoPort):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.T_supply = nan
        self.T_return = nan
        self.T_supply_id = -1
        self.T_return_id = -1
        self.dT = self.T_supply - self.T_return
        self.cp = 4200.0
        self.rho = 988.0



import logging
from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.qth_loop_mixin import QTHLoopMixin
from rtctools_heat_network.util import run_heat_network_optimization


logger = logging.getLogger("rtctools_heat_network")


class TestQTHLoop(TestCase):
    def test_basic_buffer_example(self):
        from models.basic_buffer.src.compare import (
            HeatProblemPyCML,
            QTHLoopProblemPyCML,
            base_folder,
        )

        # Just a "problem is not infeasible" test
        _heat_problem, _qth_problem = run_heat_network_optimization(
            HeatProblemPyCML, QTHLoopProblemPyCML, base_folder=base_folder
        )

    def test_double_pipe_unequal_length(self):
        import models.double_pipe_qth.src.cq2_inequality_vs_equality as cq2_inequality_vs_equality
        from models.double_pipe_qth.src.cq2_inequality_vs_equality import (
            UnequalLengthQuadraticEquality,
            UnequalLengthQuadraticEqualityLoop,
        )

        base_folder = Path(cq2_inequality_vs_equality.__file__).resolve().parent.parent

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_quadratic_equality = run_optimization_problem(
                UnequalLengthQuadraticEquality, base_folder=base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_quadratic_equality_loop = run_optimization_problem(
                UnequalLengthQuadraticEqualityLoop, base_folder=base_folder
            )

            self.assertIsInstance(equal_length_quadratic_equality_loop, QTHLoopMixin)

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        results_quadratic = equal_length_quadratic_equality.extract_results()
        results_quadratic_loop = equal_length_quadratic_equality_loop.extract_results()

        quadratic_q_1 = results_quadratic["pipe_1_hot.Q"]
        quadratic_q_2 = results_quadratic["pipe_2_hot.Q"]
        quadratic_ratio = quadratic_q_1 / quadratic_q_2

        quadratic_loop_q_1 = results_quadratic_loop["pipe_1_hot.Q"]
        quadratic_loop_q_2 = results_quadratic_loop["pipe_2_hot.Q"]
        quadratic_loop_ratio = quadratic_loop_q_1 / quadratic_loop_q_2

        # Check that both have a ratio of 2.0
        np.testing.assert_allclose(quadratic_ratio, 2.0, rtol=1e-6, atol=1e-6)
        np.testing.assert_allclose(quadratic_loop_ratio, 2.0, rtol=1e-6, atol=1e-6)



from rtctools_heat_network.pycml import Variable

from .gas_base import GasPort
from .._internal import BaseAsset
from .._internal.gas_component import GasComponent


class GasSource(GasComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_source"

        self.min_head = 30.0

        self.add_variable(GasPort, "GasOut")
        self.add_variable(Variable, "Gas_source_flow", min=0.0)

        self.add_equation((self.GasOut.Q - self.Gas_source_flow))



import json
import logging
import math
import os
from pathlib import Path
from typing import Dict, Tuple, Type, Union

import esdl
from esdl import TimeUnitEnum, UnitEnum

from rtctools_heat_network.pycml import Model as _Model

from .common import Asset
from .esdl_model_base import _RetryLaterException, _SkipAssetException

logger = logging.getLogger("rtctools_heat_network")

MODIFIERS = Dict[str, Union[str, int, float]]

HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS = 4200 * 988
WATTHOUR_TO_JOULE = 3600

MULTI_ENUM_NAME_TO_FACTOR = {
    esdl.MultiplierEnum.ATTO: 1e-18,
    esdl.MultiplierEnum.FEMTO: 1e-15,
    esdl.MultiplierEnum.PICO: 1e-12,
    esdl.MultiplierEnum.NANO: 1e-9,
    esdl.MultiplierEnum.MICRO: 1e-6,
    esdl.MultiplierEnum.MILLI: 1e-3,
    esdl.MultiplierEnum.CENTI: 1e-2,
    esdl.MultiplierEnum.DECI: 1e-1,
    esdl.MultiplierEnum.NONE: 1e0,
    esdl.MultiplierEnum.DEKA: 1e1,
    esdl.MultiplierEnum.HECTO: 1e2,
    esdl.MultiplierEnum.KILO: 1e3,
    esdl.MultiplierEnum.MEGA: 1e6,
    esdl.MultiplierEnum.GIGA: 1e9,
    esdl.MultiplierEnum.TERA: 1e12,
    esdl.MultiplierEnum.TERRA: 1e12,
    esdl.MultiplierEnum.PETA: 1e15,
    esdl.MultiplierEnum.EXA: 1e18,
}


class _AssetToComponentBase:
    # A map of pipe class name to edr asset in _edr_pipes.json
    STEEL_S1_PIPE_EDR_ASSETS = {
        "DN20": "Steel-S1-DN-20",
        "DN25": "Steel-S1-DN-25",
        "DN32": "Steel-S1-DN-32",
        "DN40": "Steel-S1-DN-40",
        "DN50": "Steel-S1-DN-50",
        "DN65": "Steel-S1-DN-65",
        "DN80": "Steel-S1-DN-80",
        "DN100": "Steel-S1-DN-100",
        "DN125": "Steel-S1-DN-125",
        "DN150": "Steel-S1-DN-150",
        "DN200": "Steel-S1-DN-200",
        "DN250": "Steel-S1-DN-250",
        "DN300": "Steel-S1-DN-300",
        "DN350": "Steel-S1-DN-350",
        "DN400": "Steel-S1-DN-400",
        "DN450": "Steel-S1-DN-450",
        "DN500": "Steel-S1-DN-500",
        "DN600": "Steel-S1-DN-600",
        "DN700": "Steel-S1-DN-700",
        "DN800": "Steel-S1-DN-800",
        "DN900": "Steel-S1-DN-900",
        "DN1000": "Steel-S1-DN-1000",
        "DN1100": "Steel-S1-DN-1100",
        "DN1200": "Steel-S1-DN-1200",
    }

    component_map = {
        "ATES": "ates",
        "ElectricityCable": "electricity_cable",
        "ElectricityDemand": "electricity_demand",
        "ElectricityProducer": "electricity_source",
        "Bus": "electricity_node",
        "GenericConsumer": "demand",
        "HeatExchange": "heat_exchanger",
        "HeatingDemand": "demand",
        "HeatPump": "heat_pump",
        "GasHeater": "source",
        "GasProducer": "gas_source",
        "GasDemand": "gas_demand",
        "GenericProducer": "source",
        "GeothermalSource": "source",
        "HeatProducer": "source",
        "ResidualHeatSource": "source",
        "GenericConversion": "heat_exchanger",
        "Joint": "node",
        "Pipe": "pipe",
        "Pump": "pump",
        "HeatStorage": "buffer",
        "Sensor": "skip",
        "Valve": "control_valve",
        "CheckValve": "check_valve",
    }

    def __init__(self):
        self._port_to_q_nominal = {}
        self._port_to_esdl_component_type = {}
        self._edr_pipes = json.load(
            open(os.path.join(Path(__file__).parent, "_edr_pipes.json"), "r")
        )

    def convert(self, asset: Asset) -> Tuple[Type[_Model], MODIFIERS]:
        """
        Converts an asset to a PyCML Heat component type and its modifiers.

        With more descriptive variable names the return type would be:
            Tuple[pycml_heat_component_type, Dict[component_attribute, new_attribute_value]]
        """
        ports = []
        if asset.in_ports is not None:
            ports.extend(asset.in_ports)
        if asset.out_ports is not None:
            ports.extend(asset.out_ports)
        assert len(ports) > 0

        for port in ports:
            self._port_to_esdl_component_type[port] = asset.asset_type

        dispatch_method_name = f"convert_{self.component_map[asset.asset_type]}"
        return getattr(self, dispatch_method_name)(asset)

    def _pipe_get_diameter_and_insulation(self, asset: Asset):
        # There are multiple ways to specify pipe properties like diameter and
        # material / insulation. We assume that DN `diameter` takes precedence
        # over `innerDiameter` and `material` (while logging warnings if both
        # are specified)
        full_name = f"{asset.asset_type} '{asset.name}'"
        if asset.attributes["innerDiameter"] and asset.attributes["diameter"].value > 0:
            logger.warning(
                f"{full_name}' has both 'innerDiameter' and 'diameter' specified. "
                f"Diameter of {asset.attributes['diameter'].name} will be used."
            )
        if asset.attributes["material"] and asset.attributes["diameter"].value > 0:
            logger.warning(
                f"{full_name}' has both 'material' and 'diameter' specified. "
                f"Insulation properties of {asset.attributes['diameter'].name} will be used."
            )
        if asset.attributes["material"] and (
            asset.attributes["diameter"].value == 0 and not asset.attributes["innerDiameter"]
        ):
            logger.warning(
                f"{full_name}' has only 'material' specified, but no information on diameter. "
                f"Diameter and insulation properties of DN200 will be used."
            )
        if asset.attributes["diameter"].value == 0 and not asset.attributes["innerDiameter"]:
            if asset.attributes["material"]:
                logger.warning(
                    f"{full_name}' has only 'material' specified, but no information on diameter. "
                    f"Diameter and insulation properties of DN200 will be used."
                )
            else:
                logger.warning(
                    f"{full_name}' has no DN size or innerDiameter specified. "
                    f"Diameter and insulation properties of DN200 will be used. "
                )

        edr_dn_size = None
        if asset.attributes["diameter"].value > 0:
            edr_dn_size = str(asset.attributes["diameter"].name)
        elif not asset.attributes["innerDiameter"]:
            edr_dn_size = "DN200"

        # NaN means the default values will be used
        insulation_thicknesses = math.nan
        conductivies_insulation = math.nan

        if edr_dn_size:
            # Get insulation and diameter properties from EDR asset with this size.
            edr_asset = self._edr_pipes[self.STEEL_S1_PIPE_EDR_ASSETS[edr_dn_size]]
            diameter = edr_asset["inner_diameter"]
            insulation_thicknesses = edr_asset["insulation_thicknesses"]
            conductivies_insulation = edr_asset["conductivies_insulation"]
        else:
            assert asset.attributes["innerDiameter"]
            diameter = asset.attributes["innerDiameter"]

            # Insulation properties
            material = asset.attributes["material"]

            if material is not None:
                if isinstance(material, esdl.esdl.MatterReference):
                    material = material.reference

                assert isinstance(material, esdl.esdl.CompoundMatter)
                components = material.component.items
                if components:
                    insulation_thicknesses = [x.layerWidth for x in components]
                    conductivies_insulation = [x.matter.thermalConductivity for x in components]

        return diameter, insulation_thicknesses, conductivies_insulation

    def _is_disconnectable_pipe(self, asset):
        # Source and buffer pipes are disconnectable by default
        assert asset.asset_type == "Pipe"
        if len(asset.in_ports) == 1 and len(asset.out_ports) == 1:
            connected_type_in = self._port_to_esdl_component_type.get(
                asset.in_ports[0].connectedTo[0], None
            )
            connected_type_out = self._port_to_esdl_component_type.get(
                asset.out_ports[0].connectedTo[0], None
            )
        else:
            raise RuntimeError("Pipe does not have 1 in port and 1 out port")
        # TODO: add other components which can be disabled and thus of which the pipes are allowed
        #  to be disabled: , "heat_exchanger", "heat_pump", "ates"
        types = {k for k, v in self.component_map.items() if v in {"source", "buffer"}}

        if types.intersection({connected_type_in, connected_type_out}):
            return True
        elif connected_type_in is None or connected_type_out is None:
            raise _RetryLaterException(
                f"Could not determine if {asset.asset_type} '{asset.name}' "
                f"is a source or buffer pipe"
            )
        else:
            return False

    def _set_q_nominal(self, asset, q_nominal):
        self._port_to_q_nominal[asset.in_ports[0]] = q_nominal
        self._port_to_q_nominal[asset.out_ports[0]] = q_nominal

    def _get_connected_q_nominal(self, asset):
        if len(asset.in_ports) == 1 and len(asset.out_ports) == 1:
            try:
                connected_port = asset.in_ports[0].connectedTo[0]
                q_nominal = self._port_to_q_nominal[connected_port]
            except KeyError:
                connected_port = asset.out_ports[0].connectedTo[0]
                q_nominal = self._port_to_q_nominal.get(connected_port, None)

            if q_nominal is not None:
                self._set_q_nominal(asset, q_nominal)
                return q_nominal
            else:
                raise _RetryLaterException(
                    f"Could not determine nominal discharge for {asset.asset_type} '{asset.name}'"
                )
        elif len(asset.in_ports) >= 2 and len(asset.out_ports) == 2:
            q_nominals = {}
            for p in asset.in_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    out_port = None
                    for p2 in asset.out_ports:
                        if p2.carrier.name.replace("_ret", "") == p.carrier.name.replace(
                            "_ret", ""
                        ):
                            out_port = p2
                    try:
                        connected_port = p.connectedTo[0]
                        q_nominal = self._port_to_q_nominal[connected_port]
                    except KeyError:
                        connected_port = out_port.connectedTo[0]
                        q_nominal = self._port_to_q_nominal.get(connected_port, None)
                    if q_nominal is not None:
                        self._port_to_q_nominal[p] = q_nominal
                        self._port_to_q_nominal[out_port] = q_nominal
                        if "_ret" in p.carrier.name:
                            q_nominals["Secondary"] = {"Q_nominal": q_nominal}
                        else:
                            q_nominals["Primary"] = {"Q_nominal": q_nominal}
                    else:
                        raise _RetryLaterException(
                            f"Could not determine nominal discharge for {asset.asset_type} "
                            f"{asset.name}"
                        )
            return q_nominals

    def _get_cost_figure_modifiers(self, asset: Asset) -> Dict:
        modifiers = {}

        if asset.attributes["costInformation"] is None:
            RuntimeWarning(f"{asset.name} has no cost information specified")
            return modifiers

        if asset.asset_type == "HeatStorage":
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.JOULE
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        elif asset.asset_type == "Pipe":
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.METRE
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        elif asset.asset_type == "HeatingDemand":
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.WATT
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)
        else:
            modifiers["variable_operational_cost_coefficient"] = self.get_variable_opex_costs(asset)
            modifiers["fixed_operational_cost_coefficient"] = self.get_fixed_opex_costs(asset)
            modifiers["investment_cost_coefficient"] = self.get_investment_costs(
                asset, per_unit=UnitEnum.WATT
            )
            modifiers["installation_cost"] = self.get_installation_costs(asset)

        return modifiers

    @staticmethod
    def _get_supply_return_temperatures(asset: Asset) -> Tuple[float, float]:
        assert len(asset.in_ports) == 1 and len(asset.out_ports) == 1
        carrier = asset.global_properties["carriers"][asset.in_ports[0].carrier.id]
        supply_temperature = carrier["supplyTemperature"]
        return_temperature = carrier["returnTemperature"]

        if carrier["__rtc_type"] == "supply":
            supply_temperature_id = carrier["id_number_mapping"]
            return_temperature_id = asset.global_properties["carriers"][
                asset.in_ports[0].carrier.id + "_ret"
            ]["id_number_mapping"]
        else:
            supply_temperature_id = asset.global_properties["carriers"][
                asset.in_ports[0].carrier.id[:-4]
            ]["id_number_mapping"]
            return_temperature_id = carrier["id_number_mapping"]

        assert supply_temperature > return_temperature
        # This is a bit dangerous, but the default (not-set) value is 0.0. We
        # however require it to be explicitly set.
        assert supply_temperature != 0.0
        assert return_temperature != 0.0

        return supply_temperature, return_temperature, supply_temperature_id, return_temperature_id

    def _supply_return_temperature_modifiers(self, asset: Asset) -> MODIFIERS:
        if len(asset.in_ports) == 1 and len(asset.out_ports) == 1:
            (
                supply_temperature,
                return_temperature,
                supply_temperature_id,
                return_temperature_id,
            ) = self._get_supply_return_temperatures(asset)
            return {
                "T_supply": supply_temperature,
                "T_return": return_temperature,
                "T_supply_id": supply_temperature_id,
                "T_return_id": return_temperature_id,
            }
        elif len(asset.in_ports) >= 2 and len(asset.out_ports) == 2:
            for p in asset.in_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    carrier = asset.global_properties["carriers"][p.carrier.id]
                    if "_ret" in p.carrier.name:
                        # This in the Secondary side carrier
                        sec_supply_temperature = carrier["supplyTemperature"]
                        sec_return_temperature_id = carrier["id_number_mapping"]
                        sec_return_temperature = carrier["returnTemperature"]
                        assert sec_supply_temperature > sec_return_temperature
                        assert sec_supply_temperature > 0.0
                    else:
                        # This in the Primary side carrier
                        prim_supply_temperature = carrier["supplyTemperature"]
                        prim_return_temperature = carrier["returnTemperature"]
                        prim_supply_temperature_id = carrier["id_number_mapping"]
                        assert prim_supply_temperature > prim_return_temperature
                        assert prim_supply_temperature > 0.0
            for p in asset.out_ports:
                if isinstance(p.carrier, esdl.HeatCommodity):
                    carrier = asset.global_properties["carriers"][p.carrier.id]
                    if "_ret" in p.carrier.name:
                        prim_return_temperature_id = carrier["id_number_mapping"]
                    else:
                        sec_supply_temperature_id = carrier["id_number_mapping"]
            if not prim_supply_temperature or not sec_supply_temperature:
                raise RuntimeError(
                    f"{asset.name} carriers are not specified correctly there should be a "
                    f"dedicated _ret carrier for each hydraulically coupled network"
                )
            temperatures = {
                "Primary": {
                    "T_supply": prim_supply_temperature,
                    "T_return": prim_return_temperature,
                    "T_supply_id": prim_supply_temperature_id,
                    "T_return_id": prim_return_temperature_id,
                },
                "Secondary": {
                    "T_supply": sec_supply_temperature,
                    "T_return": sec_return_temperature,
                    "T_supply_id": sec_supply_temperature_id,
                    "T_return_id": sec_return_temperature_id,
                },
            }
            return temperatures
        else:
            # unknown model type
            return {}

    @staticmethod
    def get_state(asset: Asset):
        if asset.attributes["state"].name == "DISABLED":
            value = 0.0
        elif asset.attributes["state"].name == "OPTIONAL":
            value = 2.0
        else:
            value = 1.0
        return value

    def get_variable_opex_costs(self, asset: Asset):
        """
        Returns the variable opex costs of an asset in Euros per Wh
        """
        cost_infos = dict()
        cost_infos["variableOperationalAndMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].variableOperationalAndMaintenanceCosts
        cost_infos["variableOperationalCosts"] = asset.attributes[
            "costInformation"
        ].variableOperationalCosts
        cost_infos["variableMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].variableMaintenanceCosts

        if all(cost_info is None for cost_info in cost_infos.values()):
            logger.warning(f"No variable OPEX cost information specified for asset {asset.name}")

        value = 0.0
        for cost_info in cost_infos.values():
            if cost_info is None:
                continue
            cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
            if unit != UnitEnum.EURO:
                logger.warning(f"Expected cost information {cost_info} to provide a cost in euros.")
                continue
            if per_time != TimeUnitEnum.NONE:
                logger.warning(
                    f"Specified OPEX for asset {asset.name} include a "
                    f"component per time, which we cannot handle."
                )
                continue
            if per_unit != UnitEnum.WATTHOUR:
                logger.warning(
                    f"Expected the specified OPEX for asset "
                    f"{asset.name} to be per Wh, but they are provided "
                    f"in {per_unit} instead."
                )
                continue
            value += cost_value

        return value

    def get_fixed_opex_costs(self, asset: Asset):
        """
        Returns the fixed opex costs of an asset in Euros per W
        """
        cost_infos = dict()
        cost_infos["fixedOperationalAndMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].fixedOperationalAndMaintenanceCosts
        cost_infos["fixedOperationalCosts"] = asset.attributes[
            "costInformation"
        ].fixedOperationalCosts
        cost_infos["fixedMaintenanceCosts"] = asset.attributes[
            "costInformation"
        ].fixedMaintenanceCosts

        if all(cost_info is None for cost_info in cost_infos.values()):
            logger.warning(f"No fixed OPEX cost information specified for asset {asset.name}")
            value = 0.0
        else:
            value = 0.0
            for cost_info in cost_infos.values():
                if cost_info is None:
                    continue
                cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
                if unit != UnitEnum.EURO:
                    RuntimeWarning(
                        f"Expected cost information {cost_info} to " f"provide a cost in euros."
                    )
                    continue
                if per_unit == UnitEnum.CUBIC_METRE:
                    # index is 0 because buffers only have one in out port
                    supply_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                        "supplyTemperature"
                    ]
                    return_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                        "returnTemperature"
                    ]
                    delta_temp = supply_temp - return_temp
                    m3_to_joule_factor = delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                    cost_value = cost_value / m3_to_joule_factor
                elif per_unit == UnitEnum.NONE:
                    if asset.asset_type == "HeatStorage":
                        size = asset.attributes["capacity"]
                        if size == 0.0:
                            # index is 0 because buffers only have one in out port
                            supply_temp = asset.global_properties["carriers"][
                                asset.in_ports[0].carrier.id
                            ]["supplyTemperature"]
                            return_temp = asset.global_properties["carriers"][
                                asset.in_ports[0].carrier.id
                            ]["returnTemperature"]
                            delta_temp = supply_temp - return_temp
                            m3_to_joule_factor = (
                                delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                            )
                            size = asset.attributes["volume"] * m3_to_joule_factor
                            if size == 0.0:
                                RuntimeWarning(f"{asset.name} has not capacity or volume set")
                                return 0.0
                    elif asset.asset_type == "ATES":
                        size = asset.attributes["maxChargeRate"]
                        if size == 0.0:
                            size = asset.attributes["capacity"] / (
                                365 * 24 * 3600 / 2
                            )  # only half a year it can load
                            if size == 0.0:
                                RuntimeWarning(
                                    f"{asset.name} has not capacity or maximum charge rate set"
                                )
                                return 0.0
                    else:
                        try:
                            size = asset.attributes["power"]
                            if size == 0.0:
                                continue
                        except KeyError:
                            return 0.0
                    cost_value = cost_value / size
                elif per_unit != UnitEnum.WATT:
                    RuntimeWarning(
                        f"Expected the specified OPEX for asset "
                        f"{asset.name} to be per W or m3, but they are provided "
                        f"in {per_unit} instead."
                    )
                    continue
                value += cost_value
        return value

    @staticmethod
    def get_cost_value_and_unit(cost_info: esdl.SingleValue):
        cost_value = cost_info.value
        unit_info = cost_info.profileQuantityAndUnit
        unit = unit_info.unit
        per_time_uni = unit_info.perTimeUnit
        per_unit = unit_info.perUnit
        multiplier = unit_info.multiplier
        per_multiplier = unit_info.perMultiplier

        cost_value *= MULTI_ENUM_NAME_TO_FACTOR[multiplier]
        cost_value /= MULTI_ENUM_NAME_TO_FACTOR[per_multiplier]

        return cost_value, unit, per_unit, per_time_uni

    def get_installation_costs(self, asset: Asset):
        cost_info = asset.attributes["costInformation"].installationCosts
        if cost_info is None:
            logger.warning(f"No installation cost info provided for asset " f"{asset.name}.")
            return 0.0
        cost_value, unit, per_unit, per_time = self.get_cost_value_and_unit(cost_info)
        if unit != UnitEnum.EURO:
            logger.warning(f"Expect cost information {cost_info} to " f"provide a cost in euros")
            return 0.0
        if not per_time == TimeUnitEnum.NONE:
            logger.warning(
                f"Specified installation costs of asset {asset.name}"
                f" include a component per time, which we "
                f"cannot handle."
            )
            return 0.0
        if not per_unit == UnitEnum.NONE:
            logger.warning(
                f"Specified installation costs of asset {asset.name}"
                f" include a component per unit {per_unit}, which we "
                f"cannot handle."
            )
            return 0.0
        return cost_value

    def get_investment_costs(self, asset: Asset, per_unit: UnitEnum = UnitEnum.WATT):
        """
        Returns the investment costs of an asset in Euros per W.
        """
        cost_info = asset.attributes["costInformation"].investmentCosts
        if cost_info is None:
            RuntimeWarning(f"No investment costs provided for asset " f"{asset.name}.")
            return 0.0
        (
            cost_value,
            unit_provided,
            per_unit_provided,
            per_time_provided,
        ) = self.get_cost_value_and_unit(cost_info)
        if unit_provided != UnitEnum.EURO:
            logger.warning(f"Expect cost information {cost_info} to " f"provide a cost in euros")
            return 0.0
        if not per_time_provided == TimeUnitEnum.NONE:
            logger.warning(
                f"Specified investment costs for asset {asset.name}"
                f" include a component per time, which we "
                f"cannot handle."
            )
            return 0.0
        if per_unit == UnitEnum.WATT:
            if not per_unit_provided == UnitEnum.WATT:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per W, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
            return cost_value
        elif per_unit == UnitEnum.WATTHOUR:
            if not per_unit_provided == UnitEnum.WATTHOUR:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per Wh, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
            return cost_value
        elif per_unit == UnitEnum.METRE:
            if not per_unit_provided == UnitEnum.METRE:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per meter, but they "
                    f"are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
            return cost_value
        elif per_unit == UnitEnum.JOULE:
            if per_unit_provided == UnitEnum.WATTHOUR:
                return cost_value / WATTHOUR_TO_JOULE
            elif per_unit_provided == UnitEnum.CUBIC_METRE:
                # index is 0 because buffers only have one in out port
                supply_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                    "supplyTemperature"
                ]
                return_temp = asset.global_properties["carriers"][asset.in_ports[0].carrier.id][
                    "returnTemperature"
                ]
                delta_temp = supply_temp - return_temp
                m3_to_joule_factor = delta_temp * HEAT_STORAGE_M3_WATER_PER_DEGREE_CELCIUS
                return cost_value / m3_to_joule_factor
            else:
                logger.warning(
                    f"Expected the specified investment costs "
                    f"of asset {asset.name} to be per Wh or m3, but "
                    f"they are provided in {per_unit_provided} "
                    f"instead."
                )
                return 0.0
        else:
            logger.warning(
                f"Cannot provide investment costs for asset " f"{asset.name} per {per_unit}"
            )
            return 0.0

    def convert_skip(self, asset: Asset):
        raise _SkipAssetException(asset)



from ._internal import HeatComponent
from .heat_port import HeatPort


class HeatTwoPort(HeatComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(HeatPort, "HeatIn")
        self.add_variable(HeatPort, "HeatOut")



from rtctools_heat_network.pycml import Variable

from ._internal import BaseAsset, HeatComponent
from .heat_port import HeatPort


class Node(HeatComponent, BaseAsset):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "node"

        self.n = 2
        assert self.n >= 2

        self.add_variable(HeatPort, "HeatConn", self.n)
        self.add_variable(Variable, "H")

        # Because the orientation of the connected pipes are important to
        # setup the heat conservation, these constraints are added in the
        # mixin.

        for i in range(1, self.n + 1):
            self.add_equation(self.HeatConn[i].H - self.H)
            # Q and Heat to be set in the mixin



import sys
from pathlib import Path
from unittest import TestCase

from rtctools.util import run_optimization_problem


class TestPipeDiameterSizingExample(TestCase):
    def test_half_network_gone(self):
        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
        from examples.pipe_diameter_sizing.src.example import (
            PipeDiameterSizingProblem,
        )  # noqa: E402, I100

        base_folder = (
            Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
        )

        del root_folder
        sys.path.pop(1)

        problem = run_optimization_problem(PipeDiameterSizingProblem, base_folder=base_folder)

        parameters = problem.parameters(0)
        diameters = {p: parameters[f"{p}.diameter"] for p in problem.hot_pipes}
        results = problem.extract_results()

        # Check that half the network is removed, i.e. 4 pipes. Note that it
        # is equally possible for the left or right side of the network to be
        # removed.
        self.assertEqual(
            len([d for d in diameters.values() if d == 0.0]),
            4,
            "More/less than 4 pipes have been removed",
        )
        # Check that the correct/specific 4 pipes on the left or 4 on the right have been removed
        pipes_removed = ["Pipe_8592", "Pipe_2927", "Pipe_9a6f", "Pipe_a718"]
        pipes_remained = ["Pipe_96bc", "Pipe_51e4", "Pipe_6b39", "Pipe_f9b0"]
        self.assertTrue(
            all(
                (elem in [k for k, d in diameters.items() if (d == 0.0)] for elem in pipes_remained)
            )
            or all(
                elem in [k for k, d in diameters.items() if (d == 0.0)] for elem in pipes_removed
            ),
            "The incorrect 4 pipes have been removed",
        )
        # Ensure that the removed pipes do not have predicted hydraulic power values
        hydraulic_power_sum = 0.0
        for pipe in diameters.keys():
            if pipe in pipes_removed:
                hydraulic_power_sum += sum(abs(results[f"{pipe}.Hydraulic_power"]))
        self.assertEqual(hydraulic_power_sum, 0.0, "Hydraulic power exists for a removed pipe")

        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation of the 3rd order function should overestimate the hydraulic
        # power when compared to the product of Q and the linear approximation of 2nd order
        # function (delta pressure).
        hydraulic_power_sum = 0.0
        hydraulic_power_post_process = 0.0
        for pipe in diameters.keys():
            if pipe in pipes_remained:
                hydraulic_power_sum += sum(abs(results[f"{pipe}.Hydraulic_power"]))
                hydraulic_power_post_process += sum(
                    abs(
                        results[f"{pipe}.dH"]
                        * parameters[f"{pipe}.rho"]
                        * 9.81
                        * results[f"{pipe}.HeatOut.Q"]
                    )
                )

        self.assertGreater(hydraulic_power_sum, hydraulic_power_post_process)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestPipeDiameterSizingExample()
    a.test_half_network_gone()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from rtctools_heat_network.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class Demand(_NonStorageComponent):
    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    Heat_in=dict(min=0.0),
                    Heat_out=dict(min=0.0),
                ),
                modifiers,
            ),
        )

        self.component_type = "demand"

        # Assumption: heat in/out and extracted is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(Variable, "Heat_demand", min=0.0, nominal=self.Heat_nominal)

        self.add_equation(
            (self.HeatOut.Heat - (self.HeatIn.Heat - self.Heat_demand)) / self.Heat_nominal
        )

        self.add_equation((self.Heat_flow - self.Heat_demand) / self.Heat_nominal)



from .grow_workflow import EndScenarioSizing, EndScenarioSizingCBC, EndScenarioSizingHIGHS
from .simulator_workflow import (
    NetworkSimulator,
    NetworkSimulatorHIGHS,
    NetworkSimulatorHIGHSTestCase,
    NetworkSimulatorHIGHSWeeklyTimeStep,
)


__all__ = [
    "EndScenarioSizing",
    "EndScenarioSizingCBC",
    "EndScenarioSizingHIGHS",
    "NetworkSimulator",
    "NetworkSimulatorHIGHS",
    "NetworkSimulatorHIGHSTestCase",
    "NetworkSimulatorHIGHSWeeklyTimeStep",
]



import logging
from pathlib import Path
from unittest import TestCase

import numpy as np

from rtctools.util import run_optimization_problem

from rtctools_heat_network.util import run_heat_network_optimization


logger = logging.getLogger("rtctools_heat_network")


class TestArtificalHeadLoss(TestCase):
    @classmethod
    def setUpClass(cls):
        import models.double_pipe_qth.src.double_pipe_qth as double_pipe_qth

        cls.base_folder = Path(double_pipe_qth.__file__).resolve().parent.parent

    def test_single_pipe(self):
        from models.double_pipe_qth.src.double_pipe_qth import SinglePipeQTH

        with self.assertLogs(logger, level="WARNING") as cm:
            run_optimization_problem(SinglePipeQTH, base_folder=self.base_folder)

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

    def test_double_pipe_unequal(self):
        from models.double_pipe_qth.src.double_pipe_qth import DoublePipeUnequalQTH

        with self.assertLogs(logger, level="WARNING") as cm:
            run_optimization_problem(DoublePipeUnequalQTH, base_folder=self.base_folder)

            self.assertIn("Pipe pipe_2_hot has artificial head loss", str(cm.output))
            self.assertIn("Pipe pipe_2_cold has artificial head loss", str(cm.output))

            self.assertNotIn("Pipe pipe_1_hot has artificial head loss", str(cm.output))
            self.assertNotIn("Pipe pipe_1_cold has artificial head loss", str(cm.output))

    def test_double_pipe_unequal_with_valve(self):
        from models.double_pipe_qth.src.double_pipe_qth import DoublePipeUnequalWithValveQTH

        with self.assertLogs(logger, level="WARNING") as cm:
            run_optimization_problem(DoublePipeUnequalWithValveQTH, base_folder=self.base_folder)

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")


class TestNoHeadLoss(TestCase):
    @classmethod
    def setUpClass(cls):
        import models.double_pipe_qth.src.double_pipe_qth as double_pipe_qth

        cls.base_folder = Path(double_pipe_qth.__file__).resolve().parent.parent

    def test_single_pipe_no_headloss(self):
        from rtctools_heat_network.qth_mixin import _MinimizeHeadLosses

        from models.double_pipe_qth.src.single_pipe_no_headloss import SinglePipeQTHNoHeadLoss

        with self.assertLogs(logger, level="WARNING") as cm:
            problem = run_optimization_problem(
                SinglePipeQTHNoHeadLoss, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

            priorities = {g.priority for g in [*problem.path_goals(), *problem.goals()]}

            self.assertNotIn(_MinimizeHeadLosses.priority, priorities)


class TestHeadLossEqualities(TestCase):
    @classmethod
    def setUpClass(cls):
        import models.double_pipe_qth.src.cq2_inequality_vs_equality as cq2_inequality_vs_equality

        cls.base_folder = Path(cq2_inequality_vs_equality.__file__).resolve().parent.parent

    def test_equal_length(self):
        from models.double_pipe_qth.src.cq2_inequality_vs_equality import (
            EqualLengthLinearEquality,
            EqualLengthQuadraticEquality,
        )

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_linear_equality = run_optimization_problem(
                EqualLengthLinearEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_quadratic_equality = run_optimization_problem(
                EqualLengthQuadraticEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        # Check results of linear head loss equation
        results_linear = equal_length_linear_equality.extract_results()

        linear_q_1 = results_linear["pipe_1_hot.Q"]
        linear_q_2 = results_linear["pipe_2_hot.Q"]
        linear_ratio = linear_q_1 / linear_q_2

        np.testing.assert_allclose(linear_ratio, 1.0)

        # Check results of quadratic head loss equation
        results_quadratic = equal_length_quadratic_equality.extract_results()

        quadratic_q_1 = results_quadratic["pipe_1_hot.Q"]
        quadratic_q_2 = results_quadratic["pipe_2_hot.Q"]
        quadratic_ratio = quadratic_q_1 / quadratic_q_2

        np.testing.assert_allclose(quadratic_ratio, 1.0)

    def test_unequal_length(self):
        from models.double_pipe_qth.src.cq2_inequality_vs_equality import (
            UnequalLengthLinearEquality,
            UnequalLengthQuadraticEquality,
        )

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_linear_equality = run_optimization_problem(
                UnequalLengthLinearEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_quadratic_equality = run_optimization_problem(
                UnequalLengthQuadraticEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        # Check results of linear head loss equation
        results_linear = equal_length_linear_equality.extract_results()

        linear_q_1 = results_linear["pipe_1_hot.Q"]
        linear_q_2 = results_linear["pipe_2_hot.Q"]
        linear_ratio = linear_q_1 / linear_q_2

        np.testing.assert_allclose(linear_ratio, 4.0)

        # Check results of quadratic head loss equation
        results_quadratic = equal_length_quadratic_equality.extract_results()

        quadratic_q_1 = results_quadratic["pipe_1_hot.Q"]
        quadratic_q_2 = results_quadratic["pipe_2_hot.Q"]
        quadratic_ratio = quadratic_q_1 / quadratic_q_2

        np.testing.assert_allclose(quadratic_ratio, 2.0)

    def test_unequal_length_valve(self):
        from models.double_pipe_qth.src.cq2_inequality_vs_equality import (
            UnequalLengthValveLinearEquality,
            UnequalLengthValveQuadraticEquality,
        )

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_linear_equality = run_optimization_problem(
                UnequalLengthValveLinearEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        with self.assertLogs(logger, level="WARNING") as cm:
            equal_length_quadratic_equality = run_optimization_problem(
                UnequalLengthValveQuadraticEquality, base_folder=self.base_folder
            )

            self.assertNotIn("artificial head loss", str(cm.output))

            if not cm.output:
                # Prevent __exit__ from failing on AssertionError
                logger.warning("Test succeeded")

        # Check results of linear head loss equation
        results_linear = equal_length_linear_equality.extract_results()

        linear_q_1 = results_linear["pipe_1_hot.Q"]
        linear_q_2 = results_linear["pipe_2_hot.Q"]
        linear_ratio = linear_q_1 / linear_q_2

        # The optimization should prefer the shorter pipe much more than would
        # follow from hydraulic equality
        np.testing.assert_array_less(10.0, linear_ratio)

        # Check results of quadratic head loss equation
        results_quadratic = equal_length_quadratic_equality.extract_results()

        quadratic_q_1 = results_quadratic["pipe_1_hot.Q"]
        quadratic_q_2 = results_quadratic["pipe_2_hot.Q"]
        quadratic_ratio = quadratic_q_1 / quadratic_q_2

        # The optimization should prefer the shorter pipe much more than would
        # follow from hydraulic equality
        np.testing.assert_array_less(10.0, quadratic_ratio)

        # Because the hydraulics (equality constraints) no longer determine
        # the results with the presence of the control valve, we expect the
        # quadratic and linear formulations to have the same result.
        np.testing.assert_allclose(linear_q_1, quadratic_q_1)
        np.testing.assert_allclose(linear_q_2, quadratic_q_2)


class TestDemandTemperatureOptions(TestCase):
    def test_demand_temperature_options(self):
        import models.basic_source_and_demand.src.qth_minimize_temperatures as qth_minimize_temperatures  # noqa: E501
        from models.basic_source_and_demand.src.qth_minimize_temperatures import (
            QTHFixedDeltaTemperature,
            QTHMinReturnMaxDeltaT,
        )

        base_folder = Path(qth_minimize_temperatures.__file__).resolve().parent.parent

        fixed_dt = run_optimization_problem(QTHFixedDeltaTemperature, base_folder=base_folder)
        min_return_max_dt = run_optimization_problem(QTHMinReturnMaxDeltaT, base_folder=base_folder)

        results_fixed_dt = fixed_dt.extract_results()
        results_min_return_max_dt = min_return_max_dt.extract_results()

        eps = 1e-4

        # Check that the minimum supply temperature is met.
        demand_tin_fixed_dt = results_fixed_dt["demand.QTHIn.T"]
        demand_tin_min_return_max_dt = results_min_return_max_dt["demand.QTHIn.T"]

        np.testing.assert_array_less(70.0 - eps, demand_tin_fixed_dt)
        np.testing.assert_array_less(70.0 - eps, demand_tin_min_return_max_dt)

        # Check that the minimum return temperature is matched for the
        # MIN_RETURN_MAX_DT case.
        demand_tout_fixed_dt = results_fixed_dt["demand.QTHOut.T"]
        demand_tout_min_return_max_dt = results_min_return_max_dt["demand.QTHOut.T"]

        np.testing.assert_array_less(45.0 - eps, demand_tout_min_return_max_dt)

        # Check that the dT is either exactly, or at most 30.0 depending on
        # the option.
        demand_dt_fixed_dt = demand_tin_fixed_dt - demand_tout_fixed_dt
        demand_dt_min_return_max_dt = demand_tin_min_return_max_dt - demand_tout_min_return_max_dt

        np.testing.assert_allclose(demand_dt_fixed_dt, 30.0, atol=eps)
        np.testing.assert_array_less(demand_dt_min_return_max_dt, 30.0 + eps)


class TestSourcesEqualTemperature(TestCase):
    def test_sources_equal_temperature(self):
        import models.unit_cases.case_2a.src.equal_temperatures as equal_temperatures
        from models.unit_cases.case_2a.src.equal_temperatures import (
            HeatProblem,
            QTHEqualTempSources,
        )

        base_folder = Path(equal_temperatures.__file__).resolve().parent.parent

        equal_temperatures = run_heat_network_optimization(
            HeatProblem, QTHEqualTempSources, base_folder=base_folder
        )
        results_eq_temp = equal_temperatures[1].extract_results()

        eps = 1e-4

        # Check that sources operate at the same temperature
        source1_temp = results_eq_temp["GeothermalSource_fafd.QTHOut.T"]
        source2_temp = results_eq_temp["GeothermalSource_27cb.QTHOut.T"]

        np.testing.assert_allclose(source1_temp, source2_temp, atol=eps)
